//
//  Copyright (C) 1993  by S.Tomabechi
//     All rights reserved
//
//   trigonometric functions
//   parameter be passed as radian
//   cos,sin,atan,acos,asin
//

#include "Fint.hpp"
#include "FintFunc.hpp"

//  tan^-1  (x)=\sum (-1)^n * x^(2n+1) /(2n+1)      -1<= x <= 1
//
//     -1            -1
//  tan  (x)=/2-tan  (1/x)
//
//     -1            -1
//  tan  (x)=/4-tan  ((1-x)/(1+x))
//
//  tan(x-y)=(tan(x)-tan(y))/(1+tan(x)*tan(y))
//
//  a=tan(y) X=tan(x)
//  x-y=x-atan(a)=atan((tan(x)-a))/(1+a*tan(x))
//  atan(X)=atan(a)+atan((X-a)/(1+a*X))
//
Fint atan(const Fint& a)
{
  if( sign(a)<0 ) return ~atan(-a) ;

  if( a>1 ) return pi(a.fdigit())/2-atan(1/a) ;

  if( a==1 ) return pi(a.fdigit())/4 ;

  Fint t(a.fdigit()+1,a.fdigit()) ;

  (t=1)>>=1 ; //   (t=1)/=2UL ;      // t=1/2
  if( a>t )  return pi(a.fdigit())/4-atan((1-a)/(1+a)) ;

  t>>=2 ;            // t=1/4
  if( a>t ) return atan(t)+atan((4*a-1)/(a+4)) ;

  t=a ;
  unsigned long n ;
  Fint s(t) ;
  Fint a2(-t*t) ;
  Fint val ;
  for( n=1 ;; ) {
    n+=2UL ;
    t*=a2 ;
    if( (val=t/n)==0 ) break ;
    s+=val ;
  }
  return s ;
}

Fint atan(const Fint& a,const Fint& b)
{
  if( b==0 ) return sign(a)*pi(a.fdigit())/2 ;
  else if( b>0 ) return atan(a/b) ;
  else if( a>=0 ) return atan(a/b)+pi(a.fdigit()) ;
  else return atan(a/b)-pi(a.fdigit()) ;
}

