//
//	Copyright (C) by S.Tomabechi 1997
//	All rights reserved.
//
//	initilaize FFT class
//
#ifdef __BORLANDC__
#pragma inline
#endif

#include "FFT.hpp"
#include "IntFuncs.hpp"
#include "Error.hpp"
#include "Mint.hpp"

#include <iostream.h>

//////////////////////////////////
//
//	initilaize FFT class
//	allocate work area and calculate primitive N-th root of 1
//
//	return value: bool	true  : succeed
//						false : fail
bool FFT::initialize(unsigned long nsize)
{
  unsigned long kk ;
  int iprime,j ;
  unsigned long p,proot ;
  unsigned long* rt ;
  logN=ilog2(nsize) ;						// bit length
											// t=logN+1
  kk=1 ;
  kk<<=(32-logN) ;							// use 3 primes > 2^31
  for( iprime=0 ; iprime<3 ; iprime++ ) {
	if( N!=0 ) delete [] roots[iprime] ;
	rt=new unsigned long[nsize+1] ;			// allocate table of roots
	roots[iprime]=rt ;

	if( !rt ) return false ;

	// search Fourier prime ( Proth type prime )
	for(;;) {
	  kk-- ;
	  p=kk*nsize+1 ;
	  if( prmdiv(p)==1 ) break ;
	}
//cout << " init " << p ;
	proot=p-1 ;								// -1 mod p	 = sqrt(1) mod p
	rt[0]=1 ;
	for( j=1 ; j<logN ; j++ ) {				// find primitive 2N-th root of unity
//if( Kronecker(proot,p)!=1 ) Error(not root) ;
	  proot=sqrt(proot,p) ;					// square root in Z/pZ
	}
	__asm {									// power of 2N-th root
	  mov esi,rt
	  mov ecx,1
	  mov edx,ecx
	}
POWER:
	__asm {
	  mov eax,edx
	  mul proot
	  div p
	  mov [ecx*4+esi],edx	// prime[j]=root^j mod p
	  inc ecx
	  cmp ecx,nsize
	  jbe POWER
	}
	inverse[iprime]=invmodp(nsize,p) ;	// N^-1 mod p
	prime[iprime]=p ;
  }

			// for Chinese remainder theorem
  crt[0]=invmodp(prime[0],prime[1]) ;  // p0^-1 mod p1
  crt[1]=invmodp(prime[0],prime[2]) ;  // p0^-1 mod p2
  crt[2]=invmodp(prime[1],prime[2]) ;  // p1^-1 mod p2

	 // lp0p1 low word of prime[0]*prime[1]
	 // hp0p1 high word of prime[0]*prime[1]
  __asm {
	mov esi,this
#ifdef __BORLANDC__
	mov eax,[esi.prime]
	mul dword ptr [4+esi.prime]
	mov [esi.lp0p1],eax
	mov [esi.hp0p1],edx
#else
	mov eax,[esi+_PRIME]
	mul dword ptr [4+esi+_PRIME]
	mov [esi+_LP0P1],eax
	mov [esi+_HP0P1],edx
#endif
  }

  N=nsize ;

#ifdef FFT_DEBUG
cout << endl << "init " << N << " " << hex << p << dec ;
cerr << endl << "init " << N << " " << hex << p << dec ;
#endif
  return true ;

}

