//
//  Copyright (C) by S.Tomabechi 1997,1999
//  All rights reserved.
//
//   Three primes FFT integer multiplication
//
//   see Joachim von zur Gathen and Juergen Gerhard
//        "Modern COmputer algebra" Cambridge University Press 1999
//          Chapter 8 Algorithm 8.25
//
//  O̑fɊւ@ FFT sŁAŌ Chinese remainder theorem
//  ɂĐmȒl߂DFourier ΐA łȂO񂨂Ȃ̂Œʏ
//  FFT xmȂDAKvƂ f[^ 2^N ́Aʏ FFT 
//   1/4 炢ōςޔ
//  FFT g߂  256,65536 Ȃǂ̏lɂ 2^32 Ƃ
//  gDꂪB̗_D
//
#ifdef __BORLANDC__
#pragma inline
#endif

#include "Mint.hpp"
#include "FFT.hpp"
#include "Error.hpp"

unsigned long* FFT::d[3]={NULL,NULL,NULL} ;
unsigned long FFT::maxN=0 ;


/////////////////////////////////////////
//
// calculate product xy=x*y
//
void FFT::fft_mul(const Mint& x,const Mint& y,Mint& xy)
{
#ifdef FFT_DEBUG
ncalled++ ;
#endif
  int iprime  ;
  unsigned long nsize ;
  unsigned long v1,v2,v3,ic,c1,c2 ;
  unsigned long p ;
  unsigned long* data ;
  unsigned long* work ;
  unsigned long* d0,*d1,*d2 ;
  unsigned long inv ;

  int xynDigit ;
  int xnDigit ;
  int ynDigit ;

//cout << "init N " << N ;

  __asm {
#ifdef __BORLANDC__

    mov  ebx,x                           // x=0 ?
    mov  eax,[ebx.nDigit]
    or   eax,eax
    jz   Zero

    mov  ebx,y                           // y=0 ?
    mov  ecx,[ebx.nDigit]
    or   ecx,ecx
    jz   Zero

    cmp  eax,ecx                        // compare digits
    jae  SetDigit
    xchg eax,ecx
                                        // exchange x and y
    mov  ebx,x
    xchg ebx,y
    mov  x,ebx

#else
    mov  ebx,x                           // x=0 ?
    mov  eax,[ebx+NDIGIT]
    or   eax,eax
    jz   Zero

    mov  ebx,y                           // y=0 ?
    mov  ecx,[ebx+NDIGIT]
    or   ecx,ecx
    jz   Zero

    cmp  eax,ecx
    jae  SetDigit
    xchg eax,ecx

    mov  ebx,x
    xchg ebx,y
    mov  x,ebx

#endif
  }
SetDigit:
  __asm {
    mov xnDigit,eax
    add eax,ecx
    mov ynDigit,ecx
    mov xynDigit,eax

    mov ebx,xy
#ifdef __BORLANDC__
    mov edx,[ebx.nDim]
#else
    mov edx,[ebx+NDIM]
#endif
    and edx,0x00ffffff
    cmp edx,eax                          // check array size
    jb  REALLOC
    jmp DEFINE_SCHEME

  }
Zero:
  __asm {
    mov  ebx,xy
    mov  dword ptr [ebx.nDigit],eax
    mov  byte ptr  [ebx.nDim+3],1
  }
  return ;

#if 0
  if( x.isZero()|| y.isZero() ) {
    xy.nDim&=_POSITIVE_ ;
    xy.nDigit=0 ;
    *xy.digit=0 ;
    return ;
  }
#endif

REALLOC:
  if( xy.ndim()<xynDigit ) {            // allocate area to save result
    if( &x==&xy || &y==&xy ) {
      xy.reallocate(xynDigit) ;
    }
    else xy.allocate(xynDigit) ;
  }

DEFINE_SCHEME:
  nsize=1 ;
  while( xnDigit>long(nsize) ) nsize<<=1 ;
  nsize<<=1 ;
  if( N!=nsize ) {             // initialize
//cout << endl << "fft change initialize " << N << "->" << nsize << " " ;
//cout << endl << "xnDigit " << xnDigit ;
    if( !initialize(nsize) ) Error(out of memory) ;
  }

  if( &x!=&y ) {
    work=new unsigned long[N+1] ;
    if( !work ) Error(allocate error) ;
  }

  if( maxN<N ) {
    if( maxN!=0 ) {
      for( iprime=0 ; iprime<3 ; iprime++ )
        delete [] d[iprime] ;
    }
    for( iprime=0 ; iprime<3 ; iprime++ ) {
      d[iprime]=new unsigned long[N+1] ;
    }
    maxN=N ;
  }

  d0=d[0] ;
  d1=d[1] ;
  d2=d[2] ;

  // FFT f[^ɕϊ
  for( iprime=0 ; iprime<3 ; iprime++ ) {
    p=prime[iprime] ;                    // assume prime > 2^31
    inv=inverse[iprime] ;
//cout << endl << "prime " << p ;
    data=d[iprime] ;
    __asm {
      mov  ecx,xnDigit // x ̌
      mov  esi,x
#ifdef __BORLANDC__
      mov    esi,[esi]
#else
      mov    esi,[esi+DIGIT]
#endif
      mov    ebx,ecx                        // digit of x
      mov    edx,p

      mov    edi,data
      lea    edi,[edi+4]

      rep movsd                             // copy digit

      mov ecx,nsize
      sub ecx,ebx
      jbe LOOP10
      xor eax,eax
      mov esi,edi
      rep stosd                             // set zero to upper digits

    }
LOOP10:
    __asm {
      mov edi,data
      lea edi,[edi+4]
    }
LOOP1:
    __asm {
      dec ebx
      jz  FFT1
      cmp    [edi+ebx*4],edx            // greater than prime ?
      jb     LOOP1
      sub    [edi+ebx*4],edx            // mod prime
      jmp LOOP1
    }
FFT1:                   // Fourier transform

    fft(data,p,roots[iprime]) ;


    if( &x!=&y ) {
      __asm {
        mov    esi,y
#ifdef __BORLANDC__
        mov    esi,[esi]
#else
        mov    esi,[esi+DIGIT]
#endif
        mov    ecx,ynDigit

        mov    ebx,ecx                        // digit of y
        mov    edx,p

        mov    edi,work
        lea    edi,[edi+4]

        rep movsd                             // copy digit

        mov ecx,nsize
        sub ecx,ebx
        jbe LOOP20
        xor eax,eax
        mov esi,edi
        rep stosd                             // set 0

      }
LOOP20:
    __asm {
      mov edi,work
      lea edi,[edi+4]
    }
LOOP2:
      __asm {
        dec ebx
        jz  FFT2
        cmp [edi+ebx*4],edx            //   greater than prime ?
        jb  LOOP2
        sub [edi+ebx*4],edx            // mod prime
        jmp LOOP2
      }
FFT2:
      fft(work,p,roots[iprime]) ;
    }
    else work=data ;

   __asm {
      mov   esi,work
      mov   edi,data
      mov   ecx,p

      mov   ebx,nsize
    }
                           // calculate convolution
Convolution:
    __asm {
      mov   eax,[ebx*4+edi]
      mul   dword ptr [ebx*4+esi]
      div   ecx
      mov   [ebx*4+edi],edx
      dec   ebx
      jnz   Convolution
    }
                          // inverse Fourier transform
    ifft(data,p,roots[iprime]) ;

    __asm {               // convert mixed radix form
//      mov   esi,1
      mov   edi,this
      mov   ebx,data
      mov esi,nsize
    }
LOOP3:
    __asm {
      mov   eax,[esi*4+ebx]   // data[i]
      mul   inv                          // data[i]*inv
      div   p                            // mod p
      mov   [esi*4+ebx],edx   // data[i]=data[i]*inv mod p
      cmp   iprime,1
      jne   PR_2                         // case iprime=1
      mov   eax,d1
      mov   eax,[esi*4+eax]   // d1[i]
      mov   edx,d0
      sub   eax,[esi*4+edx]   // d1[i]-d0[i]
      jnc   SET1
    }
ADD1:
    __asm {
#ifdef __BORLANDC__
      add   eax,[4+edi].prime
#else
      add   eax,dword ptr [4+edi+_PRIME]
#endif
    }
SET1:
    __asm {
#ifdef __BORLANDC__
      mul   dword ptr [edi].crt
      div   dword ptr [4+edi].prime
#else
      mul   dword ptr [edi+_CRT]
      div   dword ptr [4+edi+_PRIME]
#endif
      mov   ecx,d1
      mov   dword ptr [esi*4+ecx],edx
      jmp   NEXT1
    }
PR_2:
    __asm {
      cmp   iprime,2
      jne   NEXT1
      mov   eax,d2
      mov   eax,[esi*4+eax]   // d2[i]
      mov   ecx,d0
      sub   eax,[esi*4+ecx]   // d2[i]-d0[i]
      jnc   SET2
    }
ADD2:
    __asm {
#ifdef __BORLANDC__
      add   eax,dword ptr [8+edi].prime      // prime[2]
#else
      add   eax,dword ptr [8+edi+_PRIME]
#endif
    }
SET2:
    __asm {
#ifdef __BORLANDC__
      mul   dword ptr [4+edi].crt            // crt[1]
      div   dword ptr [8+edi].prime          // mod prime[2]
#else
      mul   dword ptr [4+edi+_CRT]
      div   dword ptr [8+edi+_PRIME]
#endif
      mov   eax,edx
      mov   ecx,d1
      sub   eax,[esi*4+ecx]
      jnc   SET3
    }
//ADD3:
    __asm {
#ifdef __BORLANDC__
      add   eax,dword ptr [8+edi].prime
#else
      add   eax,dword ptr [8+edi+_PRIME]
#endif
    }
SET3:
    __asm {
#ifdef __BORLANDC__
      mul   dword ptr [8+edi].crt
      div   dword ptr [8+edi].prime         // mod prime[2]
#else
      mul   dword ptr [8+edi+_CRT]
      div   dword ptr [8+edi+_PRIME]
#endif
      mov   ecx,d2
      mov   [esi*4+ecx],edx
    }
NEXT1:
   __asm {
      dec esi
      jnz LOOP3
    }
  }

        // calculate exact value using chinese remainder theorem
  __asm {
    xor esi,esi                    // clear carry
    mov c1,esi                     // c1=0
    mov c2,esi                     // c2=0
    inc esi                        // esi=1  (i)

    mov edi,this                   // edi<-this fixed

    mov ebx,xy
#ifdef __BORLANDC__
    mov ebx,[ebx]  // ebx=xy.digit fixed
#else
    mov ebx,[ebx+DIGIT]  // ebx=xy.digit fixed
#endif
  }
LOOP9:
  __asm {

    mov eax,d0
    mov ecx,[esi*4+eax]  // v1=d0[i]
    mov v1,ecx

    mov eax,d2
    mov ecx,[esi*4+eax]  // v3=d2[i]
    mov v3,ecx

    mov eax,d1
    mov eax,[esi*4+eax]  // v2=d1[i]
#ifdef __BORLANDC__
    mul dword ptr [edi.prime]       // v2*prime[0]
#else
    mul dword ptr [edi+_PRIME]     // v2*prime[0]
#endif
    add eax,v1                      // v2*prime[0]+v1
    adc edx,0                       // propagate carry
    mov v2,edx

    add c1,eax                      // add carry
    adc v2,0
#ifdef __BORLANDC__
    mov eax,[edi.lp0p1]
#else
    mov eax,[edi+_LP0P1]
#endif
    mul v3
    add eax,c1
    adc edx,0

    mov [esi*4+ebx-4],eax  // save result
    add edx,c2
    mov ic,edx

#ifdef __BORLANDC__
    mov eax,[edi.hp0p1]
#else
    mov eax,[edi+_HP0P1]
#endif
    mul v3
    add eax,ic
    adc edx,0                       // add carry
    mov c2,edx

    add eax,v2
    adc c2,0                        // save overflow
    mov c1,eax

    inc esi
    cmp esi,xynDigit
    jbe LOOP9                       // next digit

  }

  xy.nDigit=xynDigit ;
  if( xy.digit[xynDigit-1]==0 ) xy.nDigit-- ;
#ifdef __BORLANDC__
  __asm {
    mov ebx,x
    mov al,[ebx.nDim+3]
    mov ebx,y
    imul byte ptr [ebx.nDim+3]
    mov ebx,xy
    mov [ebx.nDim+3],al
  }
#else
  __asm {
    mov ebx,x
    mov al,[ebx+SIGN]
    mov ebx,y
    imul byte ptr [ebx+SIGN]
    mov ebx,xy
    mov [ebx+SIGN],al
  }
#endif


//if( N!=nsize ) Error( N!=nsize) ;
//if( maxN<20 ) Error( maxN<20) ;
//cout << " FFT " << N ;
//cout << endl << "final N " << N ;
  if( &x!=&y ) {
    delete [] work ;
  }
}
