//
//  Copyright (C) 1993  by S.Tomabechi
//     All rights reserved
//
//   trigonometric functions
//   parameter must be passed as radian
//   cos,sin,atan,acos,asin
//

#include "Fint.hpp"
#include "FintFunc.hpp"
#include "Error.hpp"

//              n  2n+1
//  sin(x)=(-1)*x    /(2n+1)!
//
//  ㋉A傫 x ɂČ덷傫,
//  x  [0,/4) ̐lɒuČvZs.
//
//                 n
//  sin(x+n)=(-1)  sin(x)
//                 n
//  cos(x+n)=(-1)  cos(x)
//
//  sin(x+/2)= cos(x)
//
//  cos(x+/2)=-sin(x)
//
//  sin(/2-x)= cos(x)
//
//  cos(/2-x)= sin(x)
//

Fint sin(const Fint& a)
{

  Fint PI=pi(a.fdigit()) ;
  Fint t(a) ;
  int _sign=1 ;          // ̕ωL^

                        // sin(t)=sin(t+2n) n : integer
  Mint n=Int(t/PI/2) ;  // t = t mod 2*
  t -= 2*n*PI ;
  if( sign(t)<0 ) t+=2*PI ;

  if( t>=PI ) { t -= PI ; _sign=-_sign ; }  // sin(t+)=-sint(t)
  if( t>PI/2 ) ~( t -= PI ) ;             // sint(-t)=sin(t)
  if( t>PI/4 ) {
    ~(t -= PI/2) ;      // t=/2-t  cos(/2-t)=sin(t) ;
    if( _sign<0 ) return ~cos(t) ;
    else return cos(t) ;
  }

  Fint tk(t) ;

  Fint s(tk) ;
  t *= t ;

  unsigned long k=1 ;

  for(;;) {
    tk*=t ;
    tk/=(k+1)*(k+2) ;
    k+=2 ;
    if( tk==0 ) break ;
    s+= ~tk ;
  }

  if( _sign<0 ) return ~s ;
  else return s ;
}

//              n  2n
//  cos(x)=(-1)*x    /(2n)!
//
//  vZ̕j, sin̏ꍇƓl
//

Fint cos(const Fint& a)
{

  Fint PI=pi(a.fdigit()) ;
  Fint t(a) ;
  int _sign=1 ;       // sign  cos ̕


  Mint n=Int(t/PI/2) ;
  t -= 2*n*PI ;
  if( sign(t)<0 ) t += 2*PI ;

  if( t >= PI ) { t -= PI ; _sign=-_sign ; }       // cos(x+)=-cos(x)
  if( t >= PI/2 ) { ~(t -= PI)  ; _sign=-_sign ; } // cos(-x)=-cos(x)
  if( t>PI/4 ) {
                                          // cos(/2-t)=sin(t)
    Fint s=sin(PI/2-t) ;
    if( _sign<0 ) return ~s ;
    else return s ;
  }


  Fint tk(t.ndigit(),t.fdigit()) ;
  tk=1 ;

  Fint s(tk) ;
  t *= t ;

  unsigned long k=0 ;
  for(;;) {
    tk *= t ;
    tk /= (k+1)*(k+2) ;
    if( tk==0 ) break ;
    s += ~tk ;
    k += 2 ;
  }
  if( _sign<0 ) return ~s ;
  else return s ;
}

//    -1       -1     QQQ
// sin  (x)=tan  (1/(1-x*x)) ;
Fint asin(const Fint& a)
{
  if( a==1 || a==-1 ) Error(Domain error) ;
  return atan(1/sqrt(1-a*a)) ;
}

//     -1            -1
//  cos  (x)=/2-sin  (x)
Fint acos(const Fint& a)
{
  return pi(a.fdigit())/2-asin(a) ;
}
