//
//  Copyright (C) 1993,1995 by S.Tomabechi
//  All rights reserved.
//
//  square root mod p
//
//    return x which satisfies x^2=a mod p
//
//  f p   a ɂ p @Ƃ a ̕ x
//
//      x^2=a mod p
//
//  ߂.
//  ̊֐𗘗pƂ, Kronecker-Jacobi L g
//  a  p Ɋւĕ]ł邱Ƃm߂Ă.
//  (a/p)=1 or 0 ̂ƂԂ. (a/p)=-1 ̂Ƃ|v邩mȂ.
//
//  1995.06.25
//    mod(p,4) ƂׂƂ mod(a,4) ȂǂƂĂ߁AASY
//    {Iɂ͐AԂl͂ł߂ł.
//
//  1995.11.03
//    p=1 mod 8 ̏ꍇ̈ԈĂ
//

#include "Mint.hpp"
#include "MBit.hpp"
#include "MintFunc.hpp"
#include "IntFuncs.hpp"
Mint sqrt_Shanks(const Mint& a,const Mint& p) ;

Mint sqrt(const Mint& a,const Mint& p)
{

  if( a.isZero() ) return a ;
                         // p = 3 mod(4)
                         // a^((p-1)/2)=(a/p)=1
                         // a*a^((p-1)/2)=a
  if( (long(p)&3)==3  ) return power(a,(p+1)>>=2,p) ;

  switch( long(p)&7 ) {    // p mod 8
    case 5 :
      if( power(a,(p-1)>>=2,p)==1 ) return power(a,(p+3)>>=3,p) ;
      else return (2*a*power(4*a,(p-5)>>=3,p))&p ;
    case 1:
      return sqrt_Shanks(a,p) ;
    case 2:
      return a ;         // p=2
    default:
      break ;
  }

 return 0 ;
}


Mint sqrt_Shanks(const Mint& a,const Mint& p)
{
  Mint q ;
  long r ;

  q=p-1 ;
  r=Val2(q) ;               // 2^r || q
  if( r>0 ) q>>=r ;

//Find_generator:
#ifdef __BORLANDC__
  Mint n(p) ;
  do {
    n=irandom((unsigned long)n) ;
  } while( Kronecker(n,p)!=-1 ) ;
#else
  Mint n(p) ;
  unsigned long nn ;
  do {
    nn=(unsigned long)n ;
    n=irandom(nn) ;
  } while( Kronecker(n,p)!=-1 ) ;
#endif

//Initialize:
  Mint x ;
  Mint y=power(n,q,p) ;

  x=power(a,(q-1)>>=1,p) ;

  Mint b ;
  b=mod(a*x*x,p) ;

  (x*=a)&=p ;
#if 0
  Mint t ;
#endif  
  Mint b2 ;
  long m ;
  long j ;
  for(;;) {
//Find_exponent:
#if 0
    if( mod(b,p)==1 ) return x ;
#else 
    if( b==1 ) return x ;    
#endif    
    b2=mod(b*b,p) ;
    for( m=1 ;; m++) {
      if( b2==1 ) break ;
      (b2*=b2)&=p;
    }
    if( m==r ) return 0 ;

//Reduce_exponent:
#if 0
    t=y ;
    for( j=r-m-1 ; j>0 ; j-- ) (t*=t)&=p ;
    y=mod(t*t,p) ;
    r=m ;
    (x*=t)&=p ;
    (b*=y)&=p ;
#else 
//    t=y ;
    for( j=r-m-1 ; j>0 ; j-- ) (y*=y)&=p ;
    (x*=y)&=p ;
    r=m ;
//    ((b*=y)*=y)&=p ;
    (((b*=y)&=p)*=y)&=p ;
#endif    
  }

}
