//
//  Copyright (C) by S.Tomabechi 1997
//  All rights reserved.
//
//  Jacobi Sum Primality Test
//
//  reference.
//    Henri Cohen
//      A Cource in Computational Algebraic Number Theory  Springer GTM 138
//
//    APRT-CL.UB  UBASIC Program Source by Akiyama,KIDA,O'Hara
//
//    return value
//        bool  :  true  N is prime
//                 false N is not prime
//
//    probably applied up to 2000 digits
//
#include "Mint.hpp"
#include "IntFuncs.hpp"
#include "MBit.hpp"
#include "TBit.hpp"
#include "Error.hpp"

const long NPmax=7 ;
const long PWmax=128 ;
const long PWmax1=PWmax-1 ;
const long NQmax=350 ;
const long Qmax=65536 ;
const long LEVELmax=18 ;
static unsigned long* ind=NULL ;
static unsigned long* fx=NULL ;
long P[NPmax+1] ;

static long Q[NQmax+1],G[NQmax+1],Inv[PWmax1+1] ;
static unsigned long* Qmem ;

static long NP[LEVELmax+1],NQ[LEVELmax+1] ;
static Mint T[LEVELmax+1] ;
static Mint JS[PWmax1+1],JW[PWmax1+1],JX[PWmax1+1] ;  // work area
static Mint J0[PWmax1+1],J1[PWmax1+1],J2[PWmax1+1] ;
static Mint S[2][PWmax1+1] ;

static const Mint* pN ;

static long LEVELnow ;
static unsigned long TestingQs ;
static unsigned long TestedQs ;
static long NQall ;
static unsigned long PK,PL,PM ;
static unsigned long PP,QQ,KK ;   // KK is index of power

void GetAllPrimes(void) ;
void GetAllQs(Mint& T) ;
void GetQsGs(Mint& T,unsigned long& TestingQs) ;
void JacobiSum(Mint* J,unsigned long a,unsigned long b) ;
static unsigned long Prmroot(unsigned long p) ;
long Vp(unsigned long X,unsigned long p) ;

Mint FinalTest(unsigned long T,Mint& et) ;
long GetPrimes2Test(Mint& t,Mint& et) ;
void GetIndexTable(unsigned long q,unsigned long g) ;
void GetFxTable(unsigned long q,unsigned long g) ;
void JNormalize(Mint* J) ;
void PrepareJacobis(Mint* J0,Mint* j1,Mint* j2) ;
void CalcJacobiPQVs(void) ;
void CalcJPQ0uJPQV(Mint* JW) ;
long MatchingRoot(Mint* J) ;
void JMul(Mint* J1J2,Mint* J1,Mint* J2) ;
void JSquare(Mint* J2,Mint* J) ;
void GetInverseTable(void) ;
void power(Mint* J,Mint& E) ;
void power(Mint* J,unsigned long E) ;
static Mint E,W,U ;

bool APRCL(const Mint& N)
{
  int lp ;
  unsigned long g ;
  bool isPrime ;
  unsigned long i,j ;
  long h ;
  Mint et ;
  Mint t ;
  unsigned long res ;
  Mint u ;


  if( N<2 ) return false ;

  pN=&N ;
  ind=new unsigned long[Qmax] ;
  fx=new unsigned long[Qmax] ;
  Qmem=ind ;

  GetAllPrimes() ;
  LEVELnow=GetPrimes2Test(t,et) ;
cout << endl << "t et " << t << " " << et ;
  if( LEVELnow==-1 ) Error( This number is too large ) ;
  if( (u=gcd(t*et,N))>1 ) {
    if( u!=N ) {
      isPrime=false ;
      goto RET ;
    }
  }

  KK=prmdiv(N) ;
//cout << endl << KK ;
  if( KK==1 ) { isPrime=true ; goto RET ; }   // N is prime
  else if( KK>1 ) { isPrime=false ; goto RET ; }   // N has divisor

  cout << endl << "Jacobi Sum Test ( APR-CL )" ;
  //
  // Main test
  //
MainStart:
  for(;;) {
    for( i=1 ; i<=NP[LEVELnow] ; i++ ) {
//cout << endl << "testing Qs " << TestingQs ;
      PP=P[i] ;
      if( t%PP!=0 ) continue ;
cout << endl << "for P=" << PP << " Q=" ;
      TestedQs=0 ;
      if( PP>=3 && power(N,Mint(PP-1),Mint(PP)*PP)!=1 ) lp=1 ;
      else lp=0 ;
LoopOnCharacter:
Try_More:
      for( j=TestedQs+1 ; j<=TestingQs ; j++ ) {
        QQ=Q[j] ;
        g=G[j] ;                     // g : primitive root
        KK=Vp(QQ-1,PP) ;             // PP^KK||QQ-1
        if( KK==0 ) continue ;       // PP dose not divide QQ-1
cout << QQ << " " ;
        PK=power(PP,KK) ;            // PK=p^k
        PL=(PP-1)*power(PP,KK-1) ;   // PL is order of group Z/(p^k)*
        PM=power(PP,KK-1) ;          // PL=PK-PM

        GetIndexTable(QQ,g) ;
        GetFxTable(QQ,g) ;

        PrepareJacobis(J0,J1,J2) ;
        CalcJacobiPQVs() ;
        CalcJPQ0uJPQV(JW) ;

        h=MatchingRoot(JW) ;
        if( h==-1 ) { isPrime=false ; goto RET ; }

        if( lp==1 ) continue ;
        if( mod(h,PP)==0 ) continue ;
        if( PP!=2 ) { lp=1 ; continue ; }
        if( KK==1 ) {
          if( (((unsigned long)N)&3)==1 )  lp=1 ;  // N=1 mod 4 ?
          continue ;
        }
        if( apower(Mint(QQ),(N-1)>>=1,N)!=-1 ) {     // Euler's criterion
          isPrime=false ;
          goto RET ;
        }
        lp=1 ;

      }

      if( lp==0 ) {
        TestedQs=TestingQs ;
        if( TestingQs<NQ[LEVELnow] ) {
          cout << endl << "retry " ;
          TestingQs++ ;
          QQ=Q[TestingQs] ;
          u=T[LEVELnow] ;
          for(;;) {
            et*=QQ ;
            res=u%QQ ;
            u/=QQ  ;
            if( res!=0 ) break ;
          }
          goto Try_More ;
        }
        break ;
      }

    }

    if( lp==1 ) break ;

    if( LEVELnow==LEVELmax ) {
      isPrime=false ;
      goto RET ;
    }

    cout << endl << "retry from beginning " ;
    LEVELnow++ ;
    t=T[LEVELnow] ;
    NP[0]=NP[LEVELnow] ;
    et=2 ;
    for( j=1 ; j<=NQ[LEVELnow] ; j++ ) {
      QQ=Q[j] ;
      u=t ;
      for(;;) {
        et*=QQ ;
        res=u%QQ ;
        u/=QQ  ;
        if(res!=0 ) break ;
      }
      if( et*et>N ) {
        TestingQs=j ;
        goto MainStart ;
      }
    }
  }


Final:
cout << endl << "final test " ;
  W=FinalTest(T[LEVELnow],et) ;
  if( W>1 && W<N ) {
    cout << endl << " find factor " ;
    isPrime=false ;
    goto RET ;
  }
  else isPrime=true ;
RET:
  if( fx ) delete [] fx ;  fx=NULL ;
  if( ind ) delete [] ind ; ind=NULL ;
  return isPrime ;

}


void GetAllPrimes(void)
{
  int j ;
//  unsigned long TT
//  long TestingQs ;
  for( j=1 ; j<=NPmax ; j++ ) P[j]=prm(j) ;
  NQall=1 ;
  Qmem[1]=2 ;

  T[0]=Mint(2*2*2*2*2*2*2) * 3*3*3*3 * 5*5*5 * 7*7 * 11*11 * 13 * 17 ;
  GetAllQs(T[0]) ;
  TestingQs=1 ; Q[1]=2 ; G[1]=1 ;

  T[1]=2*2 * 3 ;
  NP[1]=2 ;
  GetQsGs(T[1],TestingQs) ; NQ[1]=TestingQs ;

  T[2]=2*2 * 3 * 5 ;                    // *5
  NP[2]=3 ;
  GetQsGs(T[2],TestingQs) ; NQ[2]=TestingQs ;

  T[3]=2*2 * 3*3 * 5 ;                  // *3
  NP[3]=3 ;
  GetQsGs(T[3],TestingQs) ; NQ[3]=TestingQs ;

  T[4]=2*2 * 3*3 * 5 * 7;               //*7
  NP[4]=4 ;
  GetQsGs(T[4],TestingQs) ; NQ[4]=TestingQs ;

  T[5]=2*2*2 * 3*3 * 5 * 7;             // *2
  NP[5]=4 ;
  GetQsGs(T[5],TestingQs) ; NQ[5]=TestingQs ;

  T[6]=2*2*2*2 * 3*3 * 5 * 7;           // *2
  NP[6]=4 ;
  GetQsGs(T[6],TestingQs) ; NQ[6]=TestingQs ;

  T[7]=2*2*2*2 * 3*3 * 5*5 * 7;         // *5
  NP[7]=4 ;
  GetQsGs(T[7],TestingQs) ; NQ[7]=TestingQs ;

  T[8]=2*2*2*2 * 3*3*3 * 5*5 * 7  ;     // *3
  NP[8]=4 ;
  GetQsGs(T[8],TestingQs) ; NQ[8]=TestingQs ;

  T[9]=2*2*2*2*2 * 3*3*3 * 5*5 * 7  ;    // *2
  NP[9]=4 ;
  GetQsGs(T[9],TestingQs) ; NQ[9]=TestingQs ;

  T[10]=2*2*2*2*2 * 3*3*3 * 5*5 * 7*7 ;  // *7
  NP[10]=4 ;
  GetQsGs(T[10],TestingQs) ; NQ[10]=TestingQs ;


  T[11]=2*2*2*2*2*2 * 3*3*3 * 5*5 * 7*7 ;  // *2
  NP[11]=4 ;
  GetQsGs(T[11],TestingQs) ; NQ[11]=TestingQs ;

  T[12]=2*2*2*2*2*2 * 3*3*3 * 5*5 * 7*7 * 11 ;   // *11
  NP[12]=5 ;
  GetQsGs(T[12],TestingQs) ; NQ[12]=TestingQs ;

  T[13]=2*2*2*2*2*2 * 3*3*3*3 * 5*5 * 7*7 * 11 ;  // *3
  NP[13]=5 ;
  GetQsGs(T[13],TestingQs) ; NQ[13]=TestingQs ;

  T[14]=2*2*2*2*2*2 * 3*3*3*3 * 5*5 * 7*7 * 11*11 ; // *11
  NP[14]=5 ;
  GetQsGs(T[14],TestingQs) ; NQ[14]=TestingQs ;

  T[15]=Mint(2*2*2*2*2*2) * 3*3*3*3 * 5*5*5 * 7*7 * 11*11 ;  // *5
  NP[15]=5 ;
  GetQsGs(T[15],TestingQs) ; NQ[15]=TestingQs ;

  T[16]=Mint(2*2*2*2*2*2*2) * 3*3*3*3 * 5*5*5 * 7*7 * 11*11 ;   // *2
//  T[16]=2*T[15] ;
  NP[16]=5 ;
  GetQsGs(T[16],TestingQs) ; NQ[16]=TestingQs ;

  T[17]=Mint(2*2*2*2*2*2*2) * 3*3*3*3 * 5*5*5 * 7*7 * 11*11 * 13 ;   // *13
  NP[17]=6 ;
  GetQsGs(T[17],TestingQs) ; NQ[17]=TestingQs ;

                                                           // *17
  T[18]=Mint(2*2*2*2*2*2*2) * 3*3*3*3 * 5*5*5 * 7*7 * 11*11 * 13 *17;
  NP[18]=7 ;
  GetQsGs(T[18],TestingQs) ; NQ[18]=TestingQs ;

}


long GetPrimes2Test(Mint& t,Mint& et)
{
  int i,j ;
  unsigned long q,res ;
  for( i=1 ; i<=LEVELmax ; i++ ) {
    et=2 ;
    for( j=1 ; j<=NQ[i] ; j++ ) {
      q=Q[j] ;
      t=T[i] ;
      for(;;) {
        et*=q ;
        res=t%q ;
        t/=q ;
        if( res!=0 ) break ;
      }
      if( et*et>(*pN) ) {
        TestingQs=j ;
        NP[0]=NP[i] ;
        t=T[i] ;
        return i ;
      }
    }
  }
  return -1  ;
// too big number

//  LEVELnow=-1 ;
}
///////////////////////////////////////////////////////
//
//  if p3
//   J0=j(,)=^(x+f(x))
//                 1xq-2
//
//  if p=2 and k3
//   J0 is same as above
//
//   J1=j(,)*j(^2,)=^(2x+f(x))
//
//   J2=j^2(^(3*2^(k-3),^(2^(k-3))
//
void PrepareJacobis(Mint* J0,Mint* J1,Mint* J2)
{
  long j ;
  for( j=0 ; j<PK ; j++ ) { J0[j]=0 ; J1[j]=0 ; }

  if( PP>2 ) {
    JacobiSum(J0,1,1) ;     // j(,)
    return ;
  }
  if( KK==1 ) return ;

  JacobiSum(J0,1,1) ;
  if( KK==2 ) return ;

    // p=2 and k>=3
  JacobiSum(JS,2,1) ;       // j(^2,)
  JMul(J1,JS,J0) ;          // j(,)*j(^2,)

                            // j(^(3*(2^-3)),^(2^k-3))
  JacobiSum(JS,3*power(2,KK-3),power(2,KK-3)) ;
  JSquare(J2,JS) ;
}


void CalcJacobiPQVs()
{
  unsigned long j,k  ;
  unsigned long r ;     // <->
  long x ;
  unsigned long Invx ;
  unsigned long a ;   // <-> E
  int iV ;

  for( j=1 ; j<PK ; j++ ) { S[0][j]=0 ; S[1][j]=0 ; }
  S[0][0]=S[1][0]=1 ;
  r=mod((*pN),PK) ;
  GetInverseTable() ;
  if( PP!=2 ) {
    for( iV=0 ; iV<2 ; iV++ ){
      for( x=1 ; x<PK ; x++ ) {
        for( j=0 ; j<PK ; j++ ) JS[j]=J0[j] ;
        if( x%PP==0 ) continue ;     // zero divisor
        if( iV==0 ) a=x ;            // calc J^
        else {
          a=(r*x)/PK ;               // calc J^
          if( a==0 ) continue ;
        }

        power(JS,a) ;                // s1
        for( j=0 ; j<PK ; j++ ) JW[j]=0 ;
        Invx=Inv[x] ;                // x^-1 p
                                     // ЁGal(Q()/Q)
        for( j=0 ; j<PK ; j++ ) {
          k=mod((j*Invx),PK) ;
          JW[k]+=JS[j] ;
        }
        JNormalize(JW) ;
        JMul(S[iV],S[iV],JW) ;       // J^, J^
      }
    }
    return ;
  }


  // PP=2
  if( KK==1 ) {
    S[0][0]=QQ ;
    S[1][0]=1 ;
    return ;
  }
//
//  J(2,q)^2*q  q=2 k=2
//
  if( KK==2 ) {
    if( r==1 ) S[1][0]=1 ;
    for( j=0 ; j<2 ; j++ ) JS[j]=J0[j] ;
    JSquare(JS,JS) ;
    if( r==3 ) {          //  r = N mod q
      for( j=0 ; j<2 ; j++ ) {
        S[1][j]=JS[j] ;
      }
    }
    for( j=0 ; j<2 ; j++ ) S[0][j]=JS[j]*QQ ;
    return ;
  }


//
//  J(2,q,v) k>2
//

  for( iV=0 ; iV<2 ; iV++ ) {
    for( x=1 ; x<PK ; x+=2  ) {
      for( j=0 ; j<PM ; j++ ) JS[j]=J1[j] ;
      if( (x&7)==5  || (x&7)==7 ) continue ;  // =0
      if( iV==0 ) a=x ;
      else {
        a=(r*x)/PK ;
        if( a==0 ) continue ;
      }
      power(JS,a) ;
      for( j=0 ; j<PK ; j++ ) JW[j]=0 ;
      Invx=Inv[x] ;
      for( j=0 ; j<PK ; j++ ) {
        k=mod(j*Invx,PK) ;
        JW[k]+=JS[j] ;
      }
      JNormalize(JW) ;

      JMul(S[iV],S[iV],JW) ;
    }
    if( iV==0 || (r&7)==1 || (r&7)==3 ) continue ;
    JMul(S[iV],S[iV],J2) ;


  }
  return ;

}

//
//  S(p,q)=J(0,p,q)^u*J(v,p,q) ;
//
void CalcJPQ0uJPQV(Mint* JW)
{
  unsigned long i,j ;
  long k ;
  for( j=0 ; j<PK ; j++ ) JS[j]=0 ;
  for( j=0 ; j<PL ; j++ ) JS[j]=S[0][j] ;


  Mint ip=(*pN)/PK ;
  power(JS,ip) ;

  for( j=0 ; j<PK ; j++ ) JW[j]=0 ;
  for( i=0 ; i<PL ; i++ ) {
    for( j=0 ; j<PL ; j++ ) {
      k=mod(i+j,PK) ;
      JW[k]=mod(JW[k]+JS[i]*S[1][j],(*pN)) ;
    }
  }
  JNormalize(JW) ;

  for( j=0 ; j<PL ; j++ ) {
    JW[j] = mod(JW[j],*pN) ;
  }
}

//
//  check condition Lp
//
//   p^k-th root of unity  s.t. S= mod N  ?
//
long MatchingRoot(Mint* J)
{
  long i,j,k ;
  Mint W ;
  W=0 ;
  for( j=0 ; j<PL ; j++ ) W+=J[j] ;
  if( W==1 ) {
    for( j=0 ; j<PL ; j++ ) {
      if( J[j]==1 ) {              // S=^j found
        return j ;
      }
    }
  }
  else if( W!=(((*pN)-1)*(PP-1)) ) return -1 ;  // W=-p mod N
  for( i=0 ; i<PM ; i++ ) {
    if( J[i]==((*pN)-1) ) break ;    // J[i]=-1
  }
  if( i>=PM ) return -1 ;

  for( j=1 ; j<PP-1 ; j++ )  {
    k=(i+j*PM) ;
    if( J[k]!=((*pN)-1) ) {
      return -1 ;                  // not found
    }
  }
  return (i+PL) ;
}

Mint FinalTest(unsigned long t,Mint& et)
{
  unsigned long j ;
  Mint r ;
  Mint Nmodet=(*pN)%et ;
  r=1 ;
//cout << endl << "t " << t ;
  for( j=1  ; j<=t ; j++ ) {
    (r*=Nmodet)%=et ;     // 
    if( r==1 ) return r ;
    if( ((*pN)%r)==0 && r<(*pN) ) return r ;  // N is composite number
  }
  r=1 ;
  return r ;
}

//
//  set table ind[g^i mod q]=i
//  p prime  g primitive root
//
void GetIndexTable(unsigned long p,unsigned long g)
{
  long j ;
  unsigned long gp(1) ;
  for( j=1 ; j<p ; j++ ) {
    gp=mod(gp,g,p) ;                   // gp <- (gp*g) mod p
    ind[gp]=j ;                        // ind[g^i mod p]=i ;
  }
}

//
// q is prime g is primitive root
//
void GetFxTable(unsigned long q,unsigned long g)
{
  long j ;
  unsigned long gp(1) ;
  for( j=1 ; j<q-1 ; j++ ) {         // 1 x  q-2
    gp=mod(gp,g,q) ;                 // decide f(x) s.t. 1-g^x=g^f(x)
    fx[j]=ind[mod(1-long(gp),q)] ;
//    fx[j]=ind[mod(1-gp,q)] ;  //  1 f(x)  q-2
  }
}

/////////////////////////////////////////////////////////////
//
//  calculate Jacobi Sum
//
//  for p>2
//
//  j(^a,^b) = ^a(x)*^b(1-x)      q=p^k
//
//               =  ^(a*x+b*f(x))       ^q=1
//                 1<=x<=q-2
//
//  Q Z[G] ( G=<> ^PK=1 )
//  ^k ɊւēWJ,̌WvZ      0  k  PK
//
void JacobiSum(Mint* J,unsigned long a,unsigned long b)
{
  unsigned long i,j,k ;
  for( j=0 ; j<PK ; j++ ) J[j]=0 ;
  for( i=1 ; i<=QQ-2 ; i++ ) {
    j=mod(a*i+b*fx[i],PK) ;      // PK=PM+PL
    if( j<PL ) ++J[j] ;          // ^PL=^(-PM) 𗘗pď
    else {                       // wɒu
      for( k=1 ; k<PP ; k++ ) --J[j-k*PM] ;
    }
  }
}

//
//  table of inverses mod p^k
//
void GetInverseTable(void)
{
  unsigned long j ;
  for( j=1 ; j<PK ; j++ ) {
    if( j%PP ) Inv[j]=invmodp(j,PK) ;
    else Inv[j]=0 ;
  }
}


///////////////////////////////////////////////////
//
//  power of Jacobi sum
//
//  J=J^n
//
void power(Mint* J,Mint& n)
{
  long j ;
  long l=ilog2(n) ;
  MBit B ;
  if( n==1 ) return ;
  B=n ;                     //   n rbgƂĈ
  for( j=0 ; j<PL ; j++ ) JW[j]=J[j] ;
  for( j=l-1 ; j>=0 ; j-- )  {
    JSquare(J,J) ;
    if( B[j] ) {            // j-th bit on ?
      JMul(J,J,JW) ;
    }
  }
}

///////////////////////////////////////////////////
//
//  power of Jacobi sum
//
//  J=J^n
//
void power(Mint* J,unsigned long n)
{
  long j ;
  long l=ilog2(n) ;
  TBit B ;
  if( n==1 ) return ;
//  B=n ;                     //   n rbgƂĈ
  B.SetBitString(32,&n) ;
  for( j=0 ; j<PL ; j++ ) JW[j]=J[j] ;
  for( j=l-1 ; j>=0 ; j-- )  {
    JSquare(J,J) ;
    if( B[j] ) {            // j-th bit on ?
      JMul(J,J,JW) ;
    }
  }
}



//
//  product of Jacobi sum
//
//  J1J2=J1*J2
void JMul(Mint* J1J2,Mint* J1,Mint* J2)
{
  unsigned long i,j,k ;
  for( j=0 ; j<PK ; j++ ) JX[j]=0 ;

  for( i=0 ; i<PL ; i++ ) {
    for( j=0 ; j<PL ; j++ ) {
      k=mod(i+j,PK) ;
      JX[k]=mod(JX[k]+J1[i]*J2[j],*pN) ;
    }
  }
  for( j=0 ; j<PK ; j++ ) Swap(J1J2[j],JX[j]) ;
  JNormalize(J1J2) ;
}

/////////////////////////////////////////////////
//
//  square of Jacobi sum
//
//    J2=J*J
//
void JSquare(Mint* J2,Mint* J)
{
  unsigned long i,j,k ;

  for( j=0 ; j<PK ; j++ ) JX[j]=0 ;
  for( i=0 ; i<PL ; i++ ) {
    k=mod(2*i,PK) ;
    JX[k]=mod(JX[k]+J[i]*J[i],*pN) ;       // (^i)^2
  }
  for( i=0 ; i<PL ; i++ ) {
    for( j=i+1 ; j<PL ; j++ ) {
      k=mod(i+j,PK) ;
      JX[k]=mod(JX[k]+2*J[i]*J[j],*pN) ;   // (^i)*(^j)=^(i+k mod p^k)
    }
  }

  for( j=0 ; j<PK ; j++ ) Swap(J2[j],JX[j]) ;
  JNormalize(J2) ;
}


/////////////////////////////////////////////////
//
//  normalize coefficient of Jacobi sum
//
void JNormalize(Mint* J)
{
  unsigned long i,j ;
  Mint J0  ;
  for( i=PL ; i<PK ;i++ ) {
    if( J[i]!=0 ) {
      J0=J[i] ;
      for( j=1 ; j<PP ; j++ ) J[i-j*PM]-=J0 ;
      J[i]=0 ;
    }
  }
}


//
//  claculate primitive root
//
static unsigned long Prmroot(unsigned long p)
{
  unsigned long g(1),n(p-1),nn,d,sw,q,r ;
  do {
    g++ ;
    nn=n ;
    sw=1 ;
    do {
      d=prmdiv(nn) ;
      if( d==1 ) d=nn ;
      if( d>1 ) {
        for(;;) {
          q=nn/d ;
          r=nn-q*d ;
          if( r==0 ) nn=q ;
          else break ;
        }
        if( power(g,(unsigned long)(n/d),p)==1 ) { sw=0 ; nn=1 ;}
      }
    } while( nn>1 ) ;
  } while( sw==0 )  ;
  return g ;
}

//
//  find prime q s.t. (q-1)|t
//
void GetAllQs(Mint& t)
{
  unsigned long q ;
  for( q=2 ; ; ) {
    q=nextprm(q) ;
    if( t%(q-1)==0 ) {
      NQall++ ;
      Qmem[NQall]=q ;
      if( NQall>=NQmax ) break ;
    }
  }

}
//
//  find primitive root for q  s.t. (q-1)|t
//
void GetQsGs(Mint& t,unsigned long & TestingQs)
{
  long j  ;
  unsigned long q ;
  for( j=2 ; j<=NQall ; j++ ) {
    q=Qmem[j] ;
    if( q!=0 && t%(q-1)==0 ) {
      TestingQs++ ;
      Q[TestingQs]=q ;
      G[TestingQs]=Prmroot(q) ;
      Qmem[j]=0 ;
    }
  }
}

//
//  valuation
//  return index k s.t. p^k||X
//
long Vp(unsigned long X,unsigned long p)
{
  long index ;
  unsigned long res ;
  for( index=-1 ;;) {
    index++ ;
    res=X%p ;
    X/=p ;
    if( res ) return index ;
  }
}


