//
//  Copyright(C) by S.Tomabechi 2001
//  All rights reserved.
//
//  Lucase chain
//
//  Algorithm 3.5.7
//  For a sequence x_0,x_1,.. with a rule for computing x_2j from x_j
//  and a rule for computing x_{2j+1} from x_j,x_{j+1}, this algorithm
//  Computes the pair (x_n,x_{n+1}) for a given positive integer n. We
//  have n in binary as (n_0,n_1,,,,n_{B-1}) with n_{B-1} being the high-order
//  bit. We write the rules as follows x_{2j} = x_j*x_j and x{2j+1} = x_j @ x_{j+1]
//  At each step in the for() loop in the alrorithm we have u=x_j ,v=x_{j+1}
//  for some nonnegative integer j.
//
//  1. [Initialization]
//       (u,v) = (x0,x1)
//  2. [Loop]
//      for( B>j>=0 ) {
//        if( n_j==1 ) (u,v) = (u@v,v*v)
//        else          (u,v) =(u*u,u@v)
//      }
//      return (u,v);
//
//
#include "Mint.hpp"
#include "MBit.hpp"
void Lucas1(const Mint& N,Mint& P,Mint& Q,Mint n,
           Mint& U,Mint& V)
{
  Mint D=P*P-4*Q ;
  Mint a1(1),a2(0),w1(P),w2(1) ;
  Mint t ;
  unsigned long m(0) ;
  unsigned long l=len(n) ;
  Mint half=invmodp(Mint(2),N) ;
  for( ; l>0 ; l-- ) {

    if( ((unsigned long)n)&1 ) {
      t=(a1*w1+D*a2*w2) ;
      a2=(a1*w2+a2*w1)^N ;
      a1=t^N ;
      m++ ;
    }

    t=(w1*w1+D*w2*w2) ;      // ̐
    (t*=half)^=N ;
    (w2*=w1)^=N ;                 //  D ̌W
    w1=t ;
    n>>=1 ;


  }

  U=(a2*power(half,Mint(m-1),N))%N  ;
  V=(a1*power(half,Mint(m-1),N))%N  ;
}

void LucasSeqV(const Mint& P,const Mint& Q,unsigned long n,
               Mint& V)
{

  Mint Ve ;
  unsigned long n2 ;

  if( n<2 ) {
    if( n==0 ) V=2 ;
    else V=P ;
  }
  else {
    n2=n>>1 ;
    LucasSeqV(P,Q,n2,Ve) ;
    if( n&1 ) {
      Mint Vo ;
      LucasSeqV(P,Q,n2+1,Vo) ;
      V=Vo*Ve-P*power(Q,n2) ;
    }
    else {
      V=Ve*Ve-2*power(Q,n2) ;
    }
  }

}

void LucasSequence(const Mint& P,const Mint& Q,Mint& n,
                   Mint& U,Mint& V)
{
  Mint D ;
  unsigned long r ;

  Mint a1(1),a2(0),w1(P),w2(1) ;
  unsigned long m(0) ;
  unsigned long l=len(n) ;
  Mint t ;
  D=P*P-4*Q ;

  for( ; l>0 ; l-- ) {

    if( long(n)&1 ) {
      t=(a1*w1+D*a2*w2) ;
      a2=(a1*w2+a2*w1) ;
      a1=t ;

      m++ ;
    }

    t=(w1*w1+D*w2*w2) ;      // ̐
    t>>=1 ;
    w2*=w1 ;                 //  D ̌W
    w1=t ;
    n>>=1 ;

  }
  U=a2>>(m-1) ;
  V=a1>>(m-1) ;

}



// ʏ U,V ߂ɂ U=(2*v-P*u)/(P^2-4Q)
//        V=u              Ȃ񂾂ʓ|
void Lucas_chain(Mint& P,Mint& Q,Mint& N,Mint& vm,Mint& vm1)
{
    MBit bN(N) ;
Initialization:
    Mint V1(P) ;
    unsigned long V0(2) ;
//    Mint wU,wV ;
    long len=ilog2(N) ;
    Mint Pow(1) ;
//  V0=2 ;
//  V1=P ;
    vm=V0 ;
    vm1=V1 ;
//  NN=0 ;
//  bN=N ;
//  Pow=1 ;

Loop:
    for( ; len>=0 ; len-- ) {
//cout << "len " << len ;
        if( bN[len] ) {
            // compute x_{2j+1} x_j*x_{j+1}  x_{2j+2}== x_{j+1}*x_{j+1}
//              wU=u*v-power(Q,NN)*V1 ;
            vm=vm*vm1-V1*Pow ;
//              wV=v*v-V0*power(Q,NN)  ;
//              v=v*v-V0*Pow*Q   ;
            vm1=square(vm1)-V0*Pow*Q     ;
//          Pow=square(Pow) ;
            Pow*=Pow ;
            Pow*=Q ;
        }
        else {
            // compute x_2j = xj*xj x_{2j+1}=x_j*x_{j+1}
//              wV=u*v-power(Q,NN)*V1 ;
            vm1=vm*vm1-V1*Pow ;
//              wU=u*u-V0*power(Q,NN) ;
            vm=square(vm)-V0*Pow ;
//          Pow*=Pow ;
//          Pow=square(Pow) ;
            Pow*=Pow ;
        }
//      Swap(u,wU) ;
//      Swap(v,wV) ;
    }
}


void Lucas_chain(Mint& N,Mint& P,Mint& Q,Mint& n,Mint& vm,Mint& vm1)
{
    MBit bN(n) ;
    Mint V1(P) ;
    unsigned long V0(2) ;
    long len=ilog2(n) ;
    Mint Pow(1) ;
Initialization:
//    V0=2 ;
//    V1=P ;
    vm=V0 ;
    vm1=V1 ;
//    NN=0 ;
//  bN=N ;
//    Pow=1 ;

//cout << endl << "len " << len << " N " << N ;
//cout << endl << "(" << NN << "," << (NN+1) << ")" ;
//cout << endl << "((" << u << "," << v << "))" ;
Loop:
    for( ; len>=0 ; len-- ) {
        if( bN[len] ) {
//          u=power(Q,NN)
// compute x_{2j+1} x_j*x_{j+1}  x_{2j+2}== x_{j+1}*x_{j+1}
//              wU=u*v-power(Q,NN)*V1 ;
            vm=(vm*vm1-V1*Pow)^N ;
//cout << endl << u <<"*" << v <<"-" <<Pow << "*" << V1 ;
//              wV=v*v-V0*power(Q,NN)  ;
            vm1=(square(vm1)-V0*Pow*Q)^N  ;
//cout << endl << v <<"*" << v <<"-" <<Pow << "*" << V0 ;

//            NN<<=1 ;
            Pow*=Pow ;
            (Pow*=Q)^=N ;
//            ++NN ;
//cout <<    "(" << NN << "," << (NN+1) << ")" ;
//cout << "((" << wU << "," << wV << "))" ;
        }
        else {
//              wU=u*u-V0*power(Q,NN) ;
//                wU=(u*u-V0*Pow)^N ;
//cout << endl << u <<"*" << u <<"-" <<Pow << "*" << V0 ;
//              wV=u*v-power(Q,NN)*V1 ;
            vm1=(vm*vm1-V1*Pow)^N ;
            vm=(square(vm)-V0*Pow)^N ;
//cout << endl << u <<"*" << v <<"-" << Pow << "*" << V1 ;
            (Pow*=Pow)^=N ;
//            NN<<=1 ;
//cout <<    "(" << NN << "," << (NN+1) << ")" ;
//cout << "((" << wU << "," << wV << "))" ;
        }
//        Swap(u,wU) ;
//        Swap(v,wV) ;
//      u=wU ;
//      v=wV ;
//cout << endl << "  u V " << u << " " << v ;
    }
}

bool LPRP(Mint& N,Mint& P,Mint& Q)
{
    Mint D ;
    Mint A ;
    Mint M ;
    Mint B(1) ;
    Mint U,V ;


Auxiliary_parameters:
    D=P*P-4*Q ;
    if( gcd(N,2*P*Q*D)!=1 ) return false ;

    A=mod(P*P*invmodp(Q,N)-2,N) ;
    M=(N-Kronecker(D,N))>>1 ;
Binary_Lucas_chain:
    Lucas_chain(N,A,B,M,U,V) ;

Declaration:
#if 1
    if( mod(A*U-(V<<1),N)==0 ) {
        return true ; // Lucas probable prime with parameters a,b
    }
    else return false ;
#else
//Frobenius test
    if( mod(A*U-V<<1,N)!=0 ) {
        return false ;
    }
    B=power(Q,(N-1)<<1,N) ;
    if( mod(B*U,N)==2 ) {
        return true ;
    }
    else {
        return false ;
    }

#endif


}

#include <stdlib.h>
#include <iostream.h>
#include "CPUTime.hpp"
#include "Error.hpp"

#if 0
int main(void)
{
//  Mint P(1),Q(-1) ;
    Mint P(2),Q(-1) ;
    Mint u,v ;
    Mint u1,v1 ;
    Mint v2 ;
    unsigned long n ;
    unsigned long j ;
    Mint N ;
    CPUTime cpu1,cpu2 ;
    N=97 ;
    Mint NN=primorial(31)+1 ;
//  P=1 ;
//  Q=-1 ;

    P=2 ;
    Q=-1 ;
    for( j=1 ; j<20 ; j++ ) {
        N=j ;
//        N<<=1 ;
        Lucas_chain(P,Q,N,u,v) ;
        v=(2*v-P*u)/(P*P-4*Q) ;
        cout << endl << "u v " << v << "  " << u ;
        LucasSequence(P,Q,N,u1,v1) ;
        cout << endl << "u v " << u1 << "  " << v1 ;
    }
    return 1 ;

    for( j=10 ; j<1000 ; j++ ) {
        N=j ;
        cpu1.Start() ;
        Lucas_chain(P,Q,N,u,v) ;
//cout << endl << "u " << (u%NN) ;
//        Lucas_chain(NN,P,Q,N,u,v) ;
//cout << endl << "u " << (u%NN) ;
        cpu1.Stop() ;
//      cout << endl << "N " << N ;
//      cout << endl << "u " << u ;
//      cout <<  " v " << v ;
//      cout << endl << (2*v-P*u)/(P*P-4*Q) ;
        cpu2.Start() ;
        v=(2*v-P*u)/(P*P-4*Q) ;
        cout << endl << "u v " << u << " " << v ;

// LucasSeqV(P,Q,(unsigned long)N,v) ;
        LucasSequence(P,Q,N,u1,v1) ;
cout << endl << "u1 " << (u1%NN) ;
        Lucas1(NN,P,Q,N,u1,v1) ;
cout << endl << "u1 (post Lucas 1 " << (u1%NN) ;
        cpu2.Stop() ;
// cout << endl << "V " << v ;

 if( u!=u1 ) {
    cout << endl << "u  " << u ;
    cout << endl << "u1 " << u1 ;
    cout << endl << "j  " << j ;
    Error(ahoge) ;
 }
// if( v!=u1 ) Error(ahoge) ;
//      cout << endl << "u " << u ;
//      cout << endl << "v " << v ;
    }
    cout << endl << "cput1 "<< cpu1 ;
    cout << endl << "cput2 "<< cpu2 ;
//    continue ;

    v1=P ; v2=2 ;
    for( j=2 ; j<10 ; j++ ) {
        v=P*v1-Q*v2 ;
cout << endl << j<<  " " << v ;
        v2=v1 ;
        v1=v ;
    }

}
#endif