//
//	Copyright (C) by S.Tomabechi 1994-1997,1998,2000
//	All rights reserved.
//
//	multiplication of integers
//
//	use Mint::_uMtemp and Mint::_vMtemp as temporary variable
//
//	valid for even this=&a or this=&b
//
//	1997/04/25 implement Karatsuba's method
//
//	1997/06/10 implement multiplication by Fast Fourier Transformation
//
//	1997/11/09 a<0,b>0 ̂Ƃ a*=b ̌ʂ a>0 ƂȂC
//
//	1998/10/12 Karatsuba ̕@ŁA𕪊Ƃ̌
//			   2 ̙pɂȂ悤ɏCBFFT Lɗpł悤ɂ邽߁B
//
//			  ̕@́AʂȂ̂ŒP 1/2 ɂ@ɕύX
//			  (2000/01/01)
//
#ifdef __BORLANDC__
#pragma inline
#endif

#ifndef __cplusplus
#error Must use C++ for the type Mint
#endif

#include "Mint.hpp"
#include "FFT.hpp"
#include "IntFuncs.hpp"
#include "Error.hpp"

static long k,l ;

Mint& Mint::isMul(const Mint& a,const Mint& b)
{
#ifdef _MEMMAN_DEBUG
a.check(__FILE__,__LINE__) ;
b.check(__FILE__,__LINE__) ;
#endif
  int bnDigit ;
  int anDigit ;

			  // a=0 or b=0 ?
  __asm {
	mov  esi,a
	mov  edi,b
#ifdef __BORLANDC__
	mov  eax,[esi.nDigit]
	mov  ebx,[edi.nDigit]
#else
	mov  eax,[esi+NDIGIT]
	mov  ebx,[edi+NDIGIT]
#endif
	or	 eax,eax
	jz	 MultiplyZero
  }
Check_b:
   __asm {
	or	 ebx,ebx
	jz	 MultiplyZero
	cmp  eax,ebx
	jae  not_Swap_a_b
	xchg eax,ebx
	mov  a,edi
	mov  b,esi
  }
not_Swap_a_b:
  __asm {
	mov  anDigit,eax
	mov  bnDigit,ebx
	jmp  Multiply0
  }
MultiplyZero:	  // {a|>=|b|
  __asm {
	mov edi,this
#ifdef __BORLANDC__
	mov  dword ptr [edi.nDigit],0
	mov  byte ptr [edi.nDim+3],1
#else
	mov  dword ptr [edi+NDIGIT],0
	mov  byte ptr [edi+SIGN],1
#endif
	jmp  RET1
  }

Multiply0: ;
			  // multiplication of large integers by Karatsuba's method
  if( bnDigit>Karatsuba_digit ) {

	long n ;
			 // multiplication by Fast Fourier Transformation
	if( bnDigit>FFT_digit ) {
	  if( anDigit/bnDigit<2 ) {    // 傫𑵂
		n=anDigit ;
		k=1 ;
		l=0 ;
		while( n>k ) { k<<=1 ; l++ ; }	   // index such that 2^k>=n
								 // 傫Ƃ͔ɕ

//		  if( (k-n)>(n-k/2) ) {
		if( 4*k>5*n ) {
//cout << " k " << hex << k << " " << n << dec ;
		  n=k>>1 ;
		  n-- ;
		  goto KaraTsuba ;
		}
		if( l<NUM_OF_FFT_SCHEME+_FFT_BIT_ ) {
		  _fft[l-_FFT_BIT_].fft_mul(a,b,*this) ;
		  return *this ;
		}
	  }
	}
	k=n=anDigit ;
	n>>=1 ;

/**********
	__asm {
	  mov  eax,n			   // eax <- n
	  test eax,1			   // even ?
	  jz   POW2
	  inc  eax
	}
POW2:
  __asm {
	  mov edx,eax			  // save n
	  bsr ecx,eax
	  xor ebx,ebx
	  inc ebx
	  shl ebx,cl			  // the least power of 2 ebx=2^k s.t. ebx>n

	  xor edx,ebx
	  jz  SET_n 			  // n is power of 2
	  shl ebx,1
	}
SET_n:
  __asm {
	  mov n,ebx
	}
	n>>=1 ;
//	  n-- ;
*****************/
KaraTsuba:
/*************
	if( n&1 ) n++ ;
	k=n ;
	l=1 ;
	while( n!=1 ) {
	  n>>=1 ;
	  l<<=1 ;
	}
	n=l ;
	k^=n ;
	if( k==0 ) ;
	else n<<=1 ;
***********/
				 // ̕ɕ
				 // a=a1*t+a0		 t=(2^32)^n
				 // b=b1*t+b0
				 // a*b=p2*t^2+p1*t+p0
				 // p2=a1*b1 p0=a0*b0
				 // p1=(a1+a0)*(b1+b0)-p2-p0
	Mint p2(k,0),
		 p1(k,0),
		 p0(k,0) ;
	Mint a1(n,0) ;
	__asm { 			  // ̒
						  // ̊֐ł͌ exact łȂƈȂ̂
						  // ɒ
	   mov	esi,a
	   mov	ebx,n
#ifdef __BORLANDC__
	   mov	edi,[esi]
#else
	   mov	edi,[esi]
#endif
	   cmp	anDigit,ebx
	   jae	SearchZero1
	   mov	ebx,anDigit  // case a<t
	}
SearchZero1:
	__asm {
	   cmp	dword ptr [ebx*NBYTE+edi-NBYTE],0
	   jne	FindNonZero1
	   dec	ebx
	   jnz	SearchZero1
	}
FindNonZero1:
	__asm {
#ifdef __BORLANDC__
	  mov	[esi.nDigit],ebx
#else
	  mov	[esi+NDIGIT],ebx
#endif
	  mov	edi,dword ptr a1.digit

	  mov	esi,a		 // 98/01/11
#ifdef __BORLANDC__
	  mov	bl,byte ptr [esi.nDim+3]
	  mov	byte ptr [a1.nDim+3],bl
	  mov	esi,[esi]
#else
	  mov	bl,byte ptr [esi+SIGN]
	  mov	byte ptr [a1+SIGN],bl
	  mov	esi,[esi]
#endif

	  mov	ebx,n
	  lea	esi,[esi+ebx*NBYTE]
	  mov	ecx,anDigit
	  sub	ecx,ebx
	  ja	NonZero1
#ifdef __BORLANDC__
	  mov	dword ptr a1.nDigit,0	   // =0
	  mov	byte ptr [a1.nDim+3],1
#else
	  mov	dword ptr [a1+NDIGIT],0
	  mov	byte ptr [a1+SIGN],1
#endif
	  jmp	KaraMul1
	}
NonZero1:
	 __asm {
#ifdef __BORLANDC__
	  mov	dword ptr a1.nDigit,ecx
#else
	  mov	dword ptr [a1+NDIGIT],ecx
#endif
	  dec	ecx
	}
Shift1:
	__asm {
	  mov	eax,[ecx*NBYTE+esi]
	  mov	[ecx*NBYTE+edi],eax
	  dec	ecx
	  jns	Shift1
	}

KaraMul1:
	if( &a==&b ) {
	   p2.isMul(a1,a1) ;	  // p2=a1^2
	   a1.isAdd(a1,a) ; 	  // a1=a1+a0
	   p1.isMul(a1,a1) ;	  // (a1+a0)^2
	   p0.isMul(a,a) ;
	}
	else {
	  Mint b1(n,0) ;
	  __asm {
		 mov   esi,b
		 mov   ebx,n
#ifdef __BORLANDC__
		 mov   edi,[esi]
#else
		 mov   edi,[esi]
#endif
		 cmp   bnDigit,ebx
		 jae   SearchZero2
		 mov   ebx,bnDigit		 // case b<t
	   }
SearchZero2:
	  __asm {
		 cmp   dword ptr [ebx*NBYTE+edi-NBYTE],0
		 jne   FindNonZero2
		 dec   ebx
		 jnz   SearchZero2
	  }
FindNonZero2:
	  __asm {
#ifdef __BORLANDC__
		mov   [esi.nDigit],ebx
		mov   edi,dword ptr b1.digit
#else
		mov   [esi+NDIGIT],ebx
		mov   edi,dword ptr [b1]
#endif
		mov   esi,b
#ifdef __BORLANDC__
		mov   bl,byte ptr [esi.nDim+3]
		mov   byte ptr [b1.nDim+3],bl
		mov   esi,dword ptr [esi]
#else
		mov   bl,byte ptr [esi+SIGN]
		mov   byte ptr [b1+SIGN],bl
		mov   esi,dword ptr [esi]
#endif

		mov   ebx,n
		lea   esi,[esi+ebx*NBYTE]
		mov   ecx,bnDigit
		sub   ecx,ebx
		ja	  NonZero2
#ifdef __BORLANDC__
		mov   b1.nDigit,0
		mov   byte ptr [b1.nDim+3],1
#else
		mov   dword ptr [b1+NDIGIT],0
		mov   byte ptr [b1+SIGN],1
#endif
		jmp   KaraMul2
	  }
NonZero2:
	  __asm {
#ifdef __BORLANDC__
		mov   dword ptr [b1.nDigit],ecx
#else
		mov   dword ptr [b1+NDIGIT],ecx
#endif
		dec   ecx
	  }
Shift2:
	__asm {
		mov   eax,[ecx*NBYTE+esi]
		mov   [ecx*NBYTE+edi],eax
		dec   ecx
		jns   Shift2
	  }

KaraMul2:
	  p2.isMul(a1,b1) ;
	  p1.isMul(a1+=a,b1+=b ) ;
	  p0.isMul(a,b) ;
	  __asm {
		mov   eax,b
		mov   ecx,bnDigit
#ifdef __BORLANDC__
		mov   [eax.nDigit],ecx	   // Ƃ̌ɖ߂
#else
		mov   [eax+NDIGIT],ecx
#endif
	  }
	}
	__asm {
	   mov	 eax,a
	   mov	 ecx,anDigit
#ifdef __BORLANDC__
	   mov	 [eax.nDigit],ecx
#else
	   mov	 [eax+NDIGIT],ecx
#endif
	}

	(p1.isSub(p1,p2)).isSub(p1,p0) ;   // p1=p1-p2-p0

	Swap(*this,p2[2*n]) ;  // *this <- p2[2*n]

	isAdd(*this,p1[n]) ;
#ifdef _MEMMAN_DEBUG
_MemMan.check(__FILE__,__LINE__) ;
#endif
	return isAdd(*this,p0) ;
  }

  int abnDigit(anDigit+bnDigit) ; // ς̍ő包
  int Count ;
  unsigned long* bi ;
  unsigned long* pa ;
  unsigned long* pb ;
  unsigned long Carry ;
  signed char this_sign ;

#ifdef _MEMMAN_DEBUG
_MemMan.check(__FILE__,__LINE__) ;
#endif

  if( this==&a ) {		   // copy a
	_uMtemp=a ;
	pa=_uMtemp.digit ;
  }
  else pa=a.digit ;

#ifdef _MEMMAN_DEBUG
_MemMan.check(__FILE__,__LINE__) ;
#endif

  if( this==&b ) {		   // copy b
	if( this!=&a ) { _vMtemp=b ; pb=_vMtemp.digit ; }
	else pb=pa ;
  }
  else pb=b.digit ;

#ifdef _MEMMAN_DEBUG
_MemMan.check(__FILE__,__LINE__) ;
#endif

  __asm {
	mov esi,a
#ifdef __BORLANDC__
	mov al,[esi.nDim+3]
	mov esi,b
	imul byte ptr [esi.nDim+3]
#else
	mov al,byte ptr [esi+SIGN]
	mov esi,b
	imul byte ptr [esi+SIGN]
#endif
	mov this_sign,al
	mov esi,this

#ifdef __BORLANDC__
	mov eax,[esi.nDim]
#else
	mov eax,[esi+NDIM]
#endif
	and eax,_DIGITS_
	cmp eax,abnDigit
	ja	SKIP_ALLOC
  }

//	if( ndim()<=abnDigit )
  allocate(abnDigit) ;	// g

SKIP_ALLOC:
#ifdef _MEMMAN_DEBUG
_MemMan.check(__FILE__,__LINE__) ;
#endif

  __asm {
	mov   ebx,pb
	mov   esi,pa
	mov   edi,this
#ifdef __BORLANDC__
	mov   edi,[edi]
#else
	mov   edi,[edi]
#endif

	mov   bi,ebx		// save index

	mov   ecx,anDigit
	xor   eax,eax
  }
clear:					// for( j=0 ; j<ndigit ; j++ ) digit[j]=0 ;
  __asm {
	mov   [ecx*NBYTE+edi-NBYTE],eax
	dec   ecx
	jnz   clear
  }
OuterLoop:
  __asm {
	mov   ebx,bi
	mov   ecx,[ebx] 	// ecx=b.digit[j]
  }
						// case b.digit[j]!=0
InnerLoop:				// compute b.digit[j]*this
  __asm {
	mov   edx,anDigit
	mov   Count,edx
	xor   ebx,ebx
	mov   Carry,ebx
  }
Multiply:
  __asm {
	  mov	eax,[ebx*NBYTE+esi]
	  mul	ecx
	  add	eax,Carry	//	propagate carry
	  adc	edx,0		//	add carry
						//	overflow is not occured here
	  add	[ebx*NBYTE+edi],eax
	  adc	edx,0		// オ
	  mov	Carry,edx	// save oferflow
//		adc   Carry,edx
	  inc	ebx
	  dec	Count
	  jnz	Multiply
	  mov	[ebx*NBYTE+edi],edx
   }
end_InnerLoop:
  __asm {
	add 	bi,NBYTE
	lea 	edi,[edi+NBYTE]
	dec 	bnDigit
	jnz 	OuterLoop
   }

CheckDigits:
  __asm {

	mov   edi,this
	mov   al,this_sign
#ifdef __BORLANDC__
	mov   esi,[edi]
	mov   [edi.nDim+3],al
#else
	mov   esi,[edi]
	mov   [edi+SIGN],al
#endif
	mov   ebx,abnDigit
  }
IsZero: 					  // search nonzero digit
  __asm {
	cmp   dword ptr [ebx*NBYTE+esi-NBYTE],0
	jne   Set
	dec   ebx
  }
Set:
OK:
  __asm {
#ifdef __BORLANDC__
	mov [edi.nDigit],ebx
#else
	mov [edi+NDIGIT],ebx
#endif
  }

RET1:
#ifdef _MEMMAN_DEBUG
_MemMan.check(__FILE__,__LINE__) ;
check(__FILE__,__LINE__) ;
#endif
  return *this ;

}

Mint& Mint::Toom(const Mint& a,const Mint& b)
{
	int anDigit=a.nDigit ;
	int bnDigit=b.nDigit ;
	if( bnDigit<300 ) return isMul(a,b) ;
	long n=anDigit/3 ;
	Mint p1(a) ;
	Mint p22(anDigit,0) ;
	Mint p21(anDigit,0) ;
	Mint p20(anDigit,0) ;
	Mint p11(anDigit,0) ;
	Mint p10(anDigit,0) ;
	Mint p00(anDigit,0) ;

	__asm {
		mov esi,a
		mov ebx,n
#ifdef __BORLANDC__
		mov [esi.nDigit],ebx
#else
		mov [esi+NDIGIT],ebx
#endif
	}
	p1[-n] ;
	Mint p2(p1) ;
	p1.nDigit=n ;
	p2[-n] ;

	if( &a==&b ) {
	}
	else {
		Mint q1(b) ;
		__asm {
			mov esi,b
			mov ebx,n
#ifdef __BORLANDC__
			mov [esi.nDigit],ebx
#else
			mov [esi+NDIGIT],ebx
#endif
		}
		q1[-n] ;
		Mint q2(q1) ;
		q1.nDigit=n ;
		q2[-n] ;
		p22.Toom(p2,q2) ;
		p21.Toom(p2+p1,q2+q1) ;
		p20.Toom(p2+=a,q2+=b) ;
		p11.Toom(p1,q1) ;
		p10.Toom(p1+=a,q1+=b) ;
		p00.Toom(a,b) ;
		__asm {
			mov esi,b
			mov ebx,bnDigit
#ifdef __BORLANDC__
			mov [esi.nDigit],ebx
#else
			mov [esi+NDIGIT],ebx
#endif
		}

	}
	*this = p00 ;
	*this +=((p10-p11-p00)[n]) ;
	*this +=((p20+p11-p22-p00)[2*n]) ;
	*this +=((p21-p22-p11)[3*n]) ;
	*this +=(p22[4*n]) ;
	__asm {
		mov esi,a
		mov ebx,anDigit
#ifdef __BORLANDC__
		mov [esi.nDigit],ebx
#else
		mov [esi+NDIGIT],ebx
#endif
	}
	return *this ;
}
