//
//  Copyright (C) by S.Tomabechi 2000
//  All rights reserved
//
//  simple parser
//
//  Pȍ\́AԈ\Ɠ܂܌ł܂Ă܂
//
//
//   g鉉Zq
//           +,-,*,/,^
//           %,!,#
//           Zq̗Dx͕sB?
//   g֐   (啶A̋ʂ͂Ȃ)
//          prm(n) : n Ԗڂ̑f
//          Kro,Kronecker  : ]L
//          nextprm, nxtprm :
//          gcd(,,,)   : ő ́A2ȏ
//          prmdiv     : ȑfōŏ̂
//
//     u : unitary operator
//     o : binary operator
//     e : expression
//     d : integer
//     l : left parenthesis
//     r : right parebthesis
//     w : postfix operator
//     s : separator ","

#ifdef WIN32
#include <strstrea.h>
#else
#include <strstrea.h>
	#ifdef __BORLANDC__
	#include <strstrea.h>
    #else
    #include <strstream.h>
	#endif
#endif
#include <stdlib.h>
#include <algorithm>
#include <vector>
#include <iostream.h>
#include <ctype.h>
#include <string.h>
#include "Mint.hpp"
#include "Error.hpp"
using namespace std ;


#define F_PRM     (1+'Z')        // builtin function   prm
#define F_KRO     (2+'Z')        // builtin function   Kronecker
#define F_NEXTPRM     (3+'Z')    // builtin function   nextprm
#define F_GCD     (4+'Z')        // builtin function   gcd
#define F_PRMDIV     (5+'Z')     // builtin function   prmdiv
#define F_LEN     (6+'Z')     // builtin function   prmdiv
#define F_LEN10     (7+'Z')     // builtin function   prmdiv
#define F_POWMOD     (8+'Z')     // builtin function   power = x^n mod p
#define F_PHI     (9+'Z')     // cyclotomic number
#define F_APOWMOD    (10+'Z')    // builtin function   power = x^n mod p
#define F_ISQRT    (11+'Z')    // builtin function   power = x^n mod p
#define F_FIB    (12+'Z')    // builtin function   Fibonacci number = x^n mod p
#define F_LUCAS    (13+'Z')    // builtin function   Fibonacci number = x^n mod p

void Lucas_chain(Mint& P,Mint& Q,Mint& N,Mint& vm,Mint& vm1) ;

// binary tree for parser
class pars
{
public:
  char  op ;            // property of term
  pars* left ;          // left child
  pars* right ;         // right child
  pars* arg ;           // argument list
  char* exp ;           // expression of term
  pars(char* _exp,char _op):left(NULL),right(NULL),arg(NULL) {
    exp=new char[strlen(_exp)+1] ;
    strcpy(exp,_exp) ;
    op=_op ;
  }
  pars(char* _exp,pars* _left,pars* _right,char _op):arg(NULL) {
    exp=NULL ;
    left=_left ;
    right=_right ;
    op=_op ;
    exp=new char[strlen(_exp)+1] ;
    strcpy(exp,_exp) ;
  }
  pars(const pars& p) {
    exp=new char[strlen(p.exp)+1] ;
    strcpy(exp,p.exp) ;
    op=p.op ;
    left=p.left ;
    right=p.right  ;
    arg=p.arg ;
  }
  ~pars() {
    delete [] exp ;
  }
  friend ::ostream& operator<<(::ostream& os,const pars& p) ;
} ;
#ifdef PARSE_DEBUG
::ostream& operator<<(::ostream& os,const pars& p)
{
  os << p.op << " " << p.exp ;
  if( p.left!=NULL ) os << "% " << p.left->op << " " << p.left->exp ;
  if( p.right!=NULL ) os <<"% " <<  p.right->op << " " << p.right->exp ;
  return os ;
}
#endif


// compute and free memory
Mint calc(pars* p)
{
#ifdef PARSE_DEBUG
cout << endl << "calc " <<  (*p) ;
if( p->left!=NULL ) cout << endl << *(p->left) ;
if( p->right!=NULL ) cout << endl << *(p->right) ;
#endif

  Mint w1,w2 ;
  Mint w[10] ;
  int j,k ;
  char op ;
  pars* nextp ;

  if( p->op=='d' ) {         // this node is a leaf
    w1=p->exp ;              // p->exp is a integer
    delete p ;
    return w1 ;
  }
  else {
    if( p->left==NULL ) {     // unitary operator if has no left child
      op=p->exp[0] ;
      if( op=='-' ) {         // negation
        w2=calc(p->right) ;
        delete p ;
        return -w2 ;
      }
      else if( op=='+' ) {
        w2=calc(p->right) ;
        delete p ;
        return w2 ;
      }
      else {
        j=0 ;
        nextp=p->right ;
        delete p ;
        p=nextp ;
        for(;;) {
          if( p==NULL ) break ;
          nextp=(p->arg) ;
          w[j++]=calc(p) ;
//          cout << endl << "arg" << j << " " << w[j-1] ;
          p=nextp ;
        }
        if( op==F_PRM ) {   // n-th prime
          if( j!=1 ) cerr << endl << "#argument must be 1" ;
          return prm((unsigned long)w[0]) ;
        }
        else if( op==F_KRO ) {    // Kronecker symbol
          if( j!=2 ) cerr << endl << "#argument must be 2" ;
          return Kronecker(w[0],w[1]) ;
        }
        else if( op==F_NEXTPRM ) {  // next prime
          if( j!=1 ) cerr << endl << "#argument must be 1" ;
          return nextprm((unsigned long)w[0]) ;
        }
        else if( op==F_GCD ) {     // greatest common divisor
          for( k=1 ; k<j ; k++ ) {
            w[0]=gcd(w[0],w[k]) ;
          }
          return w[0] ;
        }
        else if( op==F_PRMDIV ) {   // the leaset prime divisor of integer
          if( j!=1 ) cerr << endl << "#argument must be 1" ;
          return prmdiv(w[0]) ;
        }
        else if( op==F_LEN ) {     // length as binary integer
          if( j!=1 ) cerr << endl << "#argument must be 1" ;
          return len(w[0]) ;
        }
        else if( op==F_LEN10 ) {   // length as decimal integer
          if( j!=1 ) cerr << endl << "#argument must be 1" ;
          return len10(w[0]) ;
        }
        else if( op==F_POWMOD ) {  // power modulo p
          if( j!=3 ) cerr << endl << "#argument must be 3" ;
          return power(w[0],w[1],w[2]) ;
        }
        else if( op==F_APOWMOD ) {
          if( j!=3 ) cerr << endl << "#argument must be 3" ;
          return apower(w[0],w[1],w[2]) ;
        }
        else if( op==F_PHI ) {    // cyclotomic number
          if( j!=2 ) cerr << endl << "#argument must be 2" ;
          return cyc_num((unsigned long)w[0],long(w[1])) ;
        }

        else if( op==F_ISQRT ) {    // cyclotomic number
          if( j!=1 ) cerr << endl << "#argument must be 1" ;
          return isqrt(w[0]) ;
        }
        else if( op==F_FIB ) {    // cyclotomic number
          Lucas_chain(w[1]=1,w[2]=-1,w[0],w[3],w[4]) ;
          w[0]=(2*w[4]-w[3])/5 ;
          return w[0] ;
        }
        else if( op==F_LUCAS ) {    // cyclotomic number
          Lucas_chain(w[1]=1,w[2]=-1,w[0],w[3],w[4]) ;
          return w[3] ;
        }

        else {
          cerr << endl << "unkown function " ;
          return w[0] ;
        }
      }
    }
    else if( p->right==NULL ) {  //  postfix operator
      w2=calc(p->left) ;
      op=p->exp[0] ;
      delete p ;
      if( op=='!' )               // factorial
        return factorial((unsigned long)w2) ;
      else if( op=='#' )          // primorial
        return primorial((unsigned long)w2) ;
      else {
          cerr << endl << "unkown postfix operator " ;
        return w2 ;
      }
    }
    else {                       // binary operator
      w1=calc(p->left) ;         // compute left trm
      w2=calc(p->right) ;        // compute right term
      op=p->exp[0] ;
      delete p ;
      if( op=='+' )      return w1+w2 ;                 // add
      else if( op=='-' ) return w1-w2 ;                 // subtract
      else if( op=='*' ) return w1*w2 ;                 // multiply
      else if( op=='/' ) return w1/w2 ;                 // divide
      else if( op=='%' ) return w1%w2 ;                 // remainder
      else if( op=='^' ) return power(w1,(unsigned long)w2) ; // power
      else return w2 ;
    }

  }

}

// exp (I) 
// make parse tree
// ߂č\͖؂쐬
Mint parse(const char* exp0)
{
  int j,k,lb ;
  int ip(0) ;
  int c ;
  int lexp ;
  bool change ;
  char* buf ;
  char* ex ;
  char buf2[1024] ;
  bool bdiv ;
  lexp=strlen(exp0) ;
  buf=new char[lexp+1] ;
  vector< char > a ;
  vector< pars* > vp ;
  vector< pars* > wp ;
  Mint result,w ;
  pars* p ;
  pars* q ;
  char* exp ;

  exp=new char[lexp+1] ;

  for( j=lexp-1 ; j>=0 ; j-- ) {
    if( exp0[j]!=' ' ) break ;
  }
  lexp=j+1 ;

  for( k=0,j=0 ; j<lexp ; j++ ) {    // eliminate blanks
    if( exp0[j]!=' ' ) {
        if( exp0[j]=='/' ) {
            if( j<lexp-1 ) {
                if( exp0[j+1]=='/' ) break ;
                else exp[k++]=exp0[j] ;
            }
            else exp[k++]=exp0[j] ;
        }
        else exp[k++]=exp0[j] ;
    }
  }
  exp[k]='\0' ;
//cout << endl << "exp " << exp ;
  lexp=k ;



  for( k=0,j=0 ; j<lexp ; j++ ) {
    if( exp[j]!=' ' ) k++ ;
  }
  if( k==0 ) return result=0 ;

  // 
  // analyze token
  for( lb=0,j=0 ; j<=lexp ; j++ ) { // check token
    c=exp[j] ;
    if( isdigit(c) ) buf[lb++]=c ;      //   ݂͐l̂݋
    else {
      if( lb!=0 ) {                     // lX^bNɐς
        buf[lb++]='\0' ;
        p=new pars(buf,'d') ;
        lb=0 ;
//      vp.push_back(p) ;

#if 1
        buf[0]='+' ; buf[1]='\0' ;          // integer |-> expression
        q=new pars(buf,'e') ;
        q->left=NULL ;
        q->right=p ;
        vp.push_back(q) ;
#endif

      }
      if( isalpha(c) ) {                // alphabet function name ?
        buf[lb++]=c ;
        for( j++ ; ; j++ ) {
          c=exp[j] ;
          if( isalpha(c) || isdigit(c) ) {
//            buf[lb++]=tolower(c) ;
            buf[lb++]=c ;
          }
          else {
            j-- ;
            break ;
          }
        }
        buf[lb++]='\0' ;
        strcpy(buf2,buf) ;
        for( k=0 ; k<strlen(buf2) ; k++) buf2[k]=tolower(buf2[k]) ;
        if( !strcmp(buf2,"prm") ) {
          buf[0]=F_PRM ;
          buf[1]='\0' ;
        }
        else if( !strcmp(buf2,"kronecker") || !strcmp(buf2,"kro") ) {
          buf[0]=F_KRO ;
          buf[1]='\0' ;
        }
        else if( !strcmp(buf2,"nextprm") || !strcmp(buf2,"nxtprm") ) {
          buf[0]=F_NEXTPRM ;
          buf[1]='\0' ;
        }
        else if( !strcmp(buf2,"gcd") ) {
          buf[0]=F_GCD ;
          buf[1]='\0' ;
        }
        else if( !strcmp(buf2,"prmdiv") ) {
          buf[0]=F_PRMDIV ;
          buf[1]='\0' ;
        }
        else if( !strcmp(buf2,"len") ) {
          buf[0]=F_LEN ;
          buf[1]='\0' ;
        }
        else if( !strcmp(buf2,"len10") ) {
          buf[0]=F_LEN10 ;
          buf[1]='\0' ;
        }
        else if( !strcmp(buf2,"power") ||!strcmp(buf2,"powmod")) {
          buf[0]=F_POWMOD ;
          buf[1]='\0' ;
        }
        else if( !strcmp(buf2,"apower") ||!strcmp(buf2,"apowmod")) {
          buf[0]=F_APOWMOD ;
          buf[1]='\0' ;
        }
        else if( !strcmp(buf2,"phi") || !strcmp(buf2,"cyc") ) {
          buf[0]=F_PHI ;
          buf[1]='\0' ;
        }
        else if( !strcmp(buf2,"isqrt") ) {
          buf[0]=F_ISQRT ;
          buf[1]='\0' ;
        }
        else if( !strcmp(buf2,"u") || !strcmp(buf2,"fib") ) {
          buf[0]=F_FIB ;
          buf[1]='\0' ;
        }
        else if( !strcmp(buf2,"v") || !strcmp(buf2,"lucas") ) {
          buf[0]=F_LUCAS ;
          buf[1]='\0' ;
        }
//      else break ;
        else {
            cout << endl << "unkown function " << buf << " "  ;
            break ;
        }
        p=new pars(buf,'u') ;
        lb=0 ;
        vp.push_back(p) ;

      }
      else {
        switch( c ) {
          case '+' : case '-': case '*': case '/':      // operators
          case '%' : case '!': case '#':
          case '^' :
            buf[0]=c ;
            buf[1]='\0' ;
            p=new pars(buf,'o') ;
            vp.push_back(p) ;
            break ;
          case '(':                      // parenthesis (left)
          case '[':
          case '{':
            ip++ ;
            buf[0]=c ;
            buf[1]='\0' ;
            p=new pars(buf,'l') ;
            vp.push_back(p) ;
            break ;
          case ')':                      // parenthesis (right)
          case ']':
          case '}':
            ip-- ;
            buf[0]=c ;
            buf[1]='\0' ;
            p=new pars(buf,'r') ;
            vp.push_back(p) ;
            break ;
          case ',':     // comma
            buf[0]=c ;
            buf[1]='\0' ;
            p=new pars(buf,'s') ;   // separator
            vp.push_back(p) ;
            break ;
          case ' ' :
            break ;
          case '.' : case '@': case '&': case '~': case '\\' :
            cout << "invalid character is contained" ;
            break ;
          default:     // cannot understand
            break ;
        }
      }
    }
  }

  delete [] exp ;
  if( vp.size()==0 ) return 0 ;

#ifdef PARSE_DEBUG
  for( j=0 ; j<vp.size() ; j++ ) {
    if( vp[j]->op=='d' ) Error(why does constant exist) ;
  }
#endif

#ifdef PARSE_DEBUG
cout << endl << " ###############" ;
  for( j=0 ; j<vp.size() ; j++ ) {
    cout << endl << *(vp[j]) ;
  }
#endif
#if 0
  for( j=0 ; j<vp.size()-1 ; j++ ) {
    change=false ;
    if( vp[j]->op=='o'  ) {
      if( vp[j+1]->op=='o' ) {
        for( k=j+2 ; k<vp.size() ; k++ ) {
          wp.push_back(vp[k]) ;
        }
        change=true ;
      }
      else wp.push_back(vp[j]) ;
    }
    if( change ) {
      while( !vp.empty() ) vp.pop_back() ;
      for( k=0 ; k<wp.size() ; k++ ) {
        vp.push_back(wp[k]) ;
      }
      while( !wp.empty() ) wp.pop_back() ;
    }
  }
#endif
  // ʂȂꍇ
  // KɕĂ
  if( ip>=0 ) {
    for( j=0 ; j<ip ; j++ ) {
      p=new pars(")",'r') ;
      vp.push_back(p) ;
    }
  }
  else if( ip<0 ) {
    ip=-ip ;
    for( j=0 ; j<ip ; j++ ) {
      p=new pars("(",'l') ;
      wp.push_back(p) ;
    }
    for( j=0 ; j<vp.size() ; j++ ) {
      wp.push_back(vp[j]) ;
    }
    while( !vp.empty() ) vp.pop_back() ;
    for( j=0 ; j<wp.size() ; j++ ) {
      vp.push_back(wp[j]) ;
    }
    while( !wp.empty() ) wp.pop_back() ;
  }


  for(;;) {
    change=false ;

    for( j=0 ; j<vp.size() ; j++ ) {
        if( vp[j]->op=='u' ) goto BBBB ;   // ֐܂܂Ƃ߂
    }

    // Zq ^ ̗Dx
    for( j=0 ; j<vp.size() ; j++ ) {

#if 0
      if( vp[j]->op=='o' && vp[j]->exp[0]=='^' && vp[j-1]->op=='d' && vp[j+1]->op=='d'
         || vp[j]->op=='o' && vp[j]->exp[0]=='^'  && vp[j-1]->op=='d' && vp[j+1]->op=='e'
         || vp[j]->op=='o' && vp[j]->exp[0]=='^'   && vp[j-1]->op=='e' && vp[j+1]->op=='d'
         || vp[j]->op=='o' && vp[j]->exp[0]=='^'   && vp[j-1]->op=='e' && vp[j+1]->op=='e' )
#else
      // combine expression "^" first of all
      if( vp[j]->op=='o' && vp[j]->exp[0]=='^' && vp[j-1]->op=='e' && vp[j+1]->op=='e' )
#endif
      {
        if( j==0 ) goto RET1 ;
        p=new pars(vp[j]->exp,vp[j-1],vp[j+1],'e') ;
        wp.pop_back() ;
        wp.push_back(p) ;
        for( k=j+2 ; k<vp.size() ; k++ ) {
          wp.push_back(vp[k]) ;
        }
        delete vp[j] ;
//        while( !vp.empty() ) vp.pop_back() ;
        change=true ;
        break ;
      }
      else wp.push_back(vp[j]) ;


    }
    if( change ) goto AAA ;

    while( !vp.empty() ) vp.pop_back() ;
    for( k=0 ; k<wp.size() ; k++ ) {
     vp.push_back(wp[k]) ;
    }
    while( !wp.empty() ) wp.pop_back() ;


    // postfix operator  ʏ̉Zqɂ܂Ƃ߂
    for( j=0 ; j<vp.size() ; j++ ) {
//    if( (vp[j]->exp[0]=='!' || vp[j]->exp[0]=='#' ) && j>0 && (vp[j-1]->op=='d' || vp[j-1]->op=='e')  )
      if( (vp[j]->exp[0]=='!' || vp[j]->exp[0]=='#' ) && j>0 && vp[j-1]->op=='e'    )
      {
        if( j==0 ) goto RET1 ;
        vp[j]->op='e' ;
        wp.pop_back() ;
        vp[j]->left=vp[j-1] ;
        vp[j]->right=NULL ;
        wp.push_back(vp[j]) ;
        for( k=j+1 ; k<vp.size() ; k++ ) {
          wp.push_back(vp[k]) ;
        }
        change=true ;
        break ;
      }
      else wp.push_back(vp[j]) ;


    }
    if( change ) goto AAA ;

    while( !vp.empty() ) vp.pop_back() ;
    for( k=0 ; k<wp.size() ; k++ ) {
     vp.push_back(wp[k]) ;
    }
    while( !wp.empty() ) wp.pop_back() ;

BBBB:
    // *,/  +,- Dx
    // Z͍珇ɍsȂ
    for( bdiv=false,j=0 ; j<vp.size() ; j++ ) {

        // unitary operator, expression  -> expression
      if( vp[j]->op=='u' &&  vp[j+1]->op=='e'  )     {
        vp[j]->op='e' ;
        vp[j]->left=NULL ;
        vp[j]->right=vp[j+1] ;
        wp.push_back(vp[j]) ;
        for( k=j+2 ; k<vp.size() ; k++ ) {
          wp.push_back(vp[k]) ;
        }
        change=true ;
        break ;
      }
#if 0
      else if( vp[j]->op=='o' && (vp[j]->exp[0]=='*' || vp[j]->exp[0]=='/' || vp[j]->exp[0]=='%')
            && (vp[j-1]->op=='d' ) && vp[j+1]->op=='d'
         || vp[j]->op=='o' && (vp[j]->exp[0]=='*' || vp[j]->exp[0]=='/'  || vp[j]->exp[0]=='%' )
             && (vp[j-1]->op=='d' ) && vp[j+1]->op=='e'
         || vp[j]->op=='o' && (vp[j]->exp[0]=='*' || vp[j]->exp[0]=='/' || vp[j]->exp[0]=='%' )
             && ( vp[j-1]->op=='e'  ) && vp[j+1]->op=='d'
         || vp[j]->op=='o' && (vp[j]->exp[0]=='*' || vp[j]->exp[0]=='/'  || vp[j]->exp[0]=='%')
             && (vp[j-1]->op=='e' ) && vp[j+1]->op=='e' )
#else
      // expression, operator(*,/,%), expression  -> expression
      else if( vp[j]->op=='o' && (vp[j]->exp[0]=='*' || vp[j]->exp[0]=='/' || vp[j]->exp[0]=='%')
            && vp[j-1]->op=='e'  && vp[j+1]->op=='e' )
#endif
      {
        if( j==0 ) goto RET1 ;
        if( !bdiv ) {
            p=new pars(vp[j]->exp,vp[j-1],vp[j+1],'e') ;
            wp.pop_back() ;
            wp.push_back(p) ;
            for( k=j+2 ; k<vp.size() ; k++ ) {
              wp.push_back(vp[k]) ;
            }
            delete vp[j] ;
            change=true ;
            break ;
        }
        else wp.push_back(vp[j]) ;
      }
      else wp.push_back(vp[j]) ;

      if( vp[j]->exp[0]=='/' ) bdiv=true ;
    }


    if( change ) goto AAA ;

    while( !vp.empty() ) vp.pop_back() ;
    for( k=0 ; k<wp.size() ; k++ ) {
     vp.push_back(wp[k]) ;
    }
    while( !wp.empty() ) wp.pop_back() ;


    for( j=0 ; j<vp.size() ; j++ ) {
#if 0
      if( vp[j]->op=='o' && ( vp[j]->exp[0]=='!' || vp[j]->exp[0]=='#' )
          &&  ( j==vp.size()-1 || (j!=0 && vp[j-1]->op=='d')
          || (j!=0 && vp[j-1]->op=='r')  )   )
#else
      // expression, postfix -> expression
      if( vp[j]->op=='o' && ( vp[j]->exp[0]=='!' || vp[j]->exp[0]=='#' )
          &&  ( j==vp.size()-1 || (j!=0 && vp[j-1]->op=='e')
          || (j!=0 && vp[j-1]->op=='r')  )   )
#endif
      {
        vp[j]->op='w' ;     // postfix operator
        wp.push_back(vp[j]) ;
        for( k=j+1 ; k<vp.size() ; k++ ) {
          wp.push_back(vp[k]) ;
        }
        change=true ;
        break ;

      }
      //
#if 0
      else if( vp[j]->op=='w' && ( vp[j-1]->op=='d' || vp[j-1]->op=='e' ) )  {
#else
    // expression, postfix operator -> expression
      else if( vp[j]->op=='w' &&   vp[j-1]->op=='e'  )   {
#endif
        vp[j]->op='e' ;
        wp.pop_back() ;
        vp[j]->left=vp[j-1] ;
        vp[j]->right=NULL ;
        wp.push_back(vp[j]) ;
        for( k=j+1 ; k<vp.size() ; k++ ) {
          wp.push_back(vp[k]) ;
        }
        change=true ;
        break ;
      }
      else if( vp[j]->op=='o' && ( vp[j]->exp[0]=='-' || vp[j]->exp[0]=='+' ) && vp[j+1]->op!='o'
         && (j==0 || vp[j-1]->op=='l' || vp[j-1]->op=='o' || vp[j-1]->op=='s') ) // - ̐擪ɂ
      {
        vp[j]->op='u' ;
        wp.push_back(vp[j]) ;
        for( k=j+1 ; k<vp.size() ; k++ ) {
          wp.push_back(vp[k]) ;
        }
        change=true ;
        break ;
      }
#if 0
      else if( vp[j]->op=='u' && ( vp[j+1]->op=='d' || vp[j+1]->op=='e' ) )  {
        vp[j]->op='e' ;
        vp[j]->left=NULL ;
        vp[j]->right=vp[j+1] ;
        wp.push_back(vp[j]) ;
        for( k=j+2 ; k<vp.size() ; k++ ) {
          wp.push_back(vp[k]) ;
        }
        change=true ;
        break ;
      }
#endif


#if 0
      else if( vp[j]->op=='o' && vp[j-1]->op=='d' && vp[j+1]->op=='d'
         || vp[j]->op=='o' && (vp[j-1]->op=='d' ) && vp[j+1]->op=='e'
         || vp[j]->op=='o' && vp[j-1]->op=='e' && vp[j+1]->op=='d'
         || vp[j]->op=='o' && vp[j-1]->op=='e' && vp[j+1]->op=='e' )
#else
    // expression, operator , expression  -> expression
      else if( j>=1 && vp[j]->op=='o' && vp[j-1]->op=='e' && vp[j+1]->op=='e' )
#endif
      {
        p=new pars(vp[j]->exp,vp[j-1],vp[j+1],'e') ;
        wp.pop_back() ;
        wp.push_back(p) ;
        for( k=j+2 ; k<vp.size() ; k++ ) {
          wp.push_back(vp[k]) ;
        }
        delete vp[j] ;
        change=true ;
        break ;
      }
      // ʂ菜
#if 0
      else if( vp[j]->op=='l' && ( vp[j+1]->op=='d' || vp[j+1]->op=='e' )
              && vp[j+2]->op=='r' ) {
#else
        // (expression)-> expression
      else if( vp[j]->op=='l' &&  vp[j+1]->op=='e'
              && vp[j+2]->op=='r' ) {
#endif
        wp.push_back(vp[j+1]) ;
        delete vp[j] ;
        delete vp[j+2] ;
        for( k=j+3 ; k<vp.size() ; k++ ) {
          wp.push_back(vp[k]) ;
        }
        change=true ;
        break ;
      }
#if 0
      else if( vp[j]->op=='s' && (vp[j-1]->op=='d' || vp[j-1]->op=='e'  )
              &&  (vp[j+1]->op=='d' || vp[j+1]->op=='e'  ) )
#else
    // expression,expression -> expression
      else if( vp[j]->op=='s' &&  vp[j-1]->op=='e'
              &&   vp[j+1]->op=='e' && (vp[j+2]->op=='s' || vp[j+2]->op=='r')  )
#endif

      {

        p=vp[j-1] ;
        while( p->arg!=NULL ) p=p->arg ;
        p->arg=vp[j+1] ;
        delete vp[j] ;
        for( k=j+2 ; k<vp.size() ; k++ ) {
          wp.push_back(vp[k]) ;
        }
        change=true ;
        break ;
      }
      else wp.push_back(vp[j]) ;
    }
AAA:
#ifdef PARSE_DEBUG
cout << endl << " ###############" ;
for( k=0 ; k<wp.size() ; k++ ) {
  cout << endl << *(wp[k]) ;
}
#endif

    while( !vp.empty() ) vp.pop_back() ;
    if( wp.size()!=1 ) {
       for( k=0 ; k<wp.size() ; k++ ) {
         vp.push_back(wp[k]) ;
       }
       while( !wp.empty() ) wp.pop_back() ;
       if( !change ) {
         cout << endl << "I cannot parse " ;
         break ;
       }
    }
    else break ;
  }

#ifdef PARSE_DEBUG
cout << endl << " ###############" ;
for( k=0 ; k<wp.size() ; k++ ) {
  cout << endl << *(wp[k]) ;
}
#endif
RET1:

  if( wp.size()==1 ) result=calc(wp[0]) ;
  else result=0 ;

  while( !vp.empty() ) vp.pop_back() ;
  while( !wp.empty() ) wp.pop_back() ;
  delete [] buf ;
  return result ;

}


char* parse_in(::istream& is)
{

  int ch,ch1 ;
  char* str ;
  char* wstr ;
  int k(0) ;
  int len=256 ;
  str = new char[len+1] ;
  memset(str,0x00,sizeof(char)*(len+1)) ;


  for(;;) {
    ch=is.get() ;
    if( ch=='\0' ) {
        str[k++]=(char)ch ;
        break ;
    }
    else if( ch=='\n' ) {
        str[k++]='\0' ;
        break ;
    }
    else if( ch==EOF ) {
        str[k++]='\0';
        break ;
    }
    else str[k++]=(char)ch ;
//  else str[k++]=ch ;
    if( k>len-2 ) {
        wstr=new char[len*2+1] ;
        memset(wstr,0x00,sizeof(char)*(2*len+1)) ;
        memcpy(wstr,str,k) ;
        len<<=1 ;
        delete [] str ;
        str = wstr ;
    }
  }
//  str[k]='\n' ;

  return str ;
}





#ifdef MAKE_MAIN
int main(void)
{
  Mint result ;
  char* str ;



  for(;;) {
    cout <<  "> " ;
    str=parse_in(cin) ;
    if( strlen(str)==0 ) {
        delete [] str ;
        break ;
    }
cout << endl << "str " << str ;
    result=parse(str) ;
    cout <<  str << " = " << result<< endl ;
    delete [] str ;
    if( result==0 ) break ;
  }
}
#endif
