//
//	Copyright (C) by S.Tomabechi 1999
//	All rights reserved.
//
//	Sparse Matrix class
//	intended to use for block Lanczos method
//

#ifndef _SPARSE_MATRIX_HPP_
#define _SPARSE_MATRIX_HPP_

#include <iostream.h>
#include "Intfuncs.hpp"

#ifndef __BORLANDC__
#define SM_NROW 0								  // offset of members
#define SM_NCOL 4
#define SM_MAT	8
#define SM_WROW 12
#define SM_WCOL 16
#endif

class block_vector ;
class Lanczos ;

class SparseMatrix
{
	friend class block_vector ;			// friend class
	friend class Lanczos ;
public:
	const static unsigned long isHeavy ;
	const static unsigned long isDelete ;
	const static unsigned long isLight ;
	unsigned long nrow ;
	unsigned long ncol ;
	unsigned long** mat ;
	unsigned long* w_row ;				// 1 sɊ܂܂ non-zero entry ̌
	unsigned long* w_col ;				// ͕sKv

	// transposed matrix
	SparseMatrix():nrow(0),ncol(0),mat(0),w_row(0),w_col(0)
	{
	}

	SparseMatrix(const unsigned long _nrow,const unsigned long _ncol) ;

	int SetRowVector(const unsigned long irow,const unsigned long nelem,
									 const unsigned long* index) ;

	~SparseMatrix() ;
	void SGauss(void) ;
	unsigned long DeleteColumn(unsigned long* new_col) ;
	void SparseMatrix::statics(::ostream& os) const ;
	unsigned long memory(void) const ;
	unsigned long NumberOfNonZeroEntry(void) const ;
	double entry_per_row(void) const ;
	unsigned long DeleteSingleton(unsigned long* old_row) ;
	void SparseMatrix::DeleteNullRow(unsigned long* old_row) ;
	void SparseMatrix::DeleteRow(unsigned long* old_row) ;
	friend void write(::ostream& os,const SparseMatrix& m) ;
	friend void read(::istream& is,SparseMatrix& m,
				const unsigned long maxrow,const unsigned long maxcol) ;
	friend void readTransposed(::istream& is,SparseMatrix& m,
				const unsigned long maxrow,const unsigned long maxcol) ;


	friend void Mul(const unsigned long* v,const SparseMatrix& m,unsigned long* vm) ;
	friend void Mul(const SparseMatrix& m,const unsigned long* v,unsigned long* mv) ;

#if 0
	SparseMatrix(const SparseMatrix& m) ;
	friend	SparseMatrix t(const SparseMatrix& m) ;
	SparseMatrix& operator=(const SparseMatrix& m) ;
	SparseMatrix  operator*(const SparseMatrix& m) ;
	friend ::ostream& operator<<(::ostream& os,const SparseMatrix& m) ;
#endif
} ;

inline unsigned long SparseMatrix::memory(void) const
{
	return (NumberOfNonZeroEntry()*4+1023)/1024 ;
}

inline double SparseMatrix::entry_per_row(void) const
{
	return ((double)NumberOfNonZeroEntry())/nrow ;
}

#endif