//
//  Copyright (C) by S.Tomabechi 2001
//  All rights reserved.
//
//  Initializing Sieving Table
//
//  Compute  a^-1 mod p
//           2*B_j/a mod p
//           (B_0+...+B_{n-1})/a mod p
//   a is the coefficient of quadratic polynomial ax^2+2bx+c
//
//

#ifdef __BORLANDC__
#pragma inline
#endif

#define _ONLY_MINTDECL_
#include "Mint.hpp"
#include "PPMPQS.hpp"


extern unsigned long SieveWidth ;
extern unsigned long SieveUnit ;

extern char lg1[] ;
extern unsigned long ip[] ;
extern unsigned long pq[] ;
extern Mint BS[] ;


// a^-1 mod p vZ
// 2*b/a mod p vZ ̒l Bmodp[][*] ɃZbg
//  (B[0]+B[1]+...+B[n])/a mod p ͌vZ

void SetTable(Mint& A,unsigned long ncomb,unsigned long* Bmodp[],
        unsigned long nPrimes,PPrimes* Primes)
{
    unsigned long a ;
    unsigned long q_index ;
    unsigned long k(0) ;
    unsigned long r ;
    unsigned long v3,t3 ;
    unsigned long bsum ;
    PPrimes* Paddr ;

    __asm {
        mov edi,2           // edi is a prime of factor base
        mov q_index,1
        mov esi,Primes      // address of factor bases
        add esi,40          // omitt -1 and 2
        sub nPrimes,2
    }
MAIN_LOOP:
    __asm {
        movzx eax,word ptr [esi]    // difference of primes
        add   edi,eax
        dec   nPrimes
        js    RET

        inc   q_index

#if 0                               // compute a mod p
        mov   ebx,ncomb
        dec   ebx
        mov   eax,[pq+4*ebx]
        xor   edx,edx
        div   edi
        mov   ecx,edx
    }
PRO_START:              // compute a=q_1*q_2*...*q_s
    __asm {
        dec   ebx
        js    PRO_END
        mov   eax,[pq+4*ebx]
//      xor   edx,edx
        mul   ecx
        div   edi
        mov   ecx,edx
        jmp   PRO_START
    }
#else                               // ̕ ?
        mov  ebx,A
        mov  ecx,[ebx.nDigit]
        mov  ebx,[ebx.digit]
        dec  ecx
        xor  edx,edx
    }
DIV_START:
    __asm {
        mov eax,[ebx+4*ecx]
        div edi
        dec ecx
        jns DIV_START

    }
#endif

PRO_END:
    __asm {
        or   edx,edx
        jnz  SET_PARM

        mov  ebx,k      // in this case a=0 mod q i.e q is one of q_1,..,q_s
        mov  al,byte ptr [esi+2]        // logP Ҕ
        mov  byte ptr [lg1+ebx],al
#ifdef LENEAR_SIEVE
        shr  byte ptr [esi+2],1         // <- logP/2  in liear sieve case
                                        // ̂ŁA2*(1/2) Œ͍
#else
        mov  [esi+2],dl                 // logP <- 0
        mov  byte ptr [esi.dmy],dl      // ̑f p | a ł邱Ƃ
        mov  [esi+8],edx
        mov  [esi+12],edx
#endif
//      mov  [esi+16],edx
        lea  esi,[esi+20]       // 20=sizeof(PPrimes)
        inc  k                  // remove q from sieve primes
        jmp   MAIN_LOOP
    }
SET_PARM:
    __asm {

        xor ebx,ebx                     // compute a^-1 mod p

        mov   v3,edi
        mov   ecx,1                         // ecx is inverse
        mov   eax,edx
      }
DIV1:
    __asm {
        xor   edx,edx
        div   v3                            // a/v3
        mov   t3,edx                        // save residue
        mul   ebx
        sub   ecx,eax                       // ecx -= q*v1
        xchg  ecx,ebx
        mov   eax,t3
        or    eax,eax
        jz    RET1
        xchg  eax,v3
        jmp   DIV1
    }

RET1:
    __asm {
        add ecx,edi
        cmp ecx,edi
        jbe BELOW
        sub ecx,edi
    }
BELOW:
    __asm {
        mov a,ecx           // a= a^-1 mod q
        mov [esi+16],ecx    // Primes[].ainv=a^-1 mod q

        mov  Paddr,esi

        xor esi,esi
        mov bsum,esi
    }
NEXT_B:                         
    __asm {
        mov ebx,esi
        shl ebx,4
        lea ebx,[BS+ebx]        // B[q_index]
        mov ecx,[ebx.nDigit]
        mov ebx,[ebx.digit]
        xor edx,edx
        dec ecx
    }
DIV_B:
    __asm {
        mov eax,[ebx+4*ecx]     //  compute BS[j] % q
        div edi
        dec ecx
        jns DIV_B

        mov ecx,edx

        mov eax,edx
        shl eax,1           // *2
        mul a               // 2*B_j/a mod q
        div edi

        mov ebx,Bmodp           // unsigned long* Bmodp[]
        lea ebx,[ebx+4*esi]
        mov ebx,[ebx]
        mov eax,q_index
        mov [ebx+4*eax],edx     // Bmodp[j][]=2*B[j]/a mod p

        mov edx,bsum
        add edx,ecx
        sub edx,edi
        jnc LESS
        add edx,edi
    }
LESS:
    __asm {
        mov bsum,edx                  // (B[0]+...B[j]) mod q

        inc esi
        cmp esi,ncomb
        jb  NEXT_B
        mov ebx,Bmodp
        lea ebx,[ebx+4*esi]
        mov ebx,[ebx]
        mov ecx,q_index
        mov [ebx+4*ecx],edx        // (B[0]+...B[ncomb-1])   mod q

        mov esi,Paddr
        lea esi,[esi+20]
        jmp MAIN_LOOP

    }
RET: ;

}

